# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""collection of utility functions"""

import datetime as dt
import json
import random
import re
import string
from calendar import month_abbr, timegm
from collections.abc import Iterator
from textwrap import wrap

import icalendar
import pytz
import urwid
from click import style

from .parse_datetime import guesstimedeltafstr
from .terminal import get_color


def generate_random_uid() -> str:
    """generate a random uid

    when random isn't broken, getting a random UID from a pool of roughly 10^56
    should be good enough"""
    choice = string.ascii_uppercase + string.digits
    return "".join([random.choice(choice) for _ in range(36)])


RESET = "\x1b[0m"

ansi_reset = re.compile(r"\x1b\[0m")
ansi_sgr = re.compile(
    r"\x1b\["
    "(?!0m)"  # negative lookahead, don't match 0m
    "([0-9]+;?)+"
    "m"
)


def find_last_reset(string: str) -> tuple[int, int, str]:
    last = None
    for m in re.finditer(ansi_reset, string):
        last = m
    if last is None:
        return -2, -1, ""
    return last.start(), last.end(), last.group(0)


def find_last_sgr(string: str) -> tuple[int, int, str]:
    last = None
    for m in re.finditer(ansi_sgr, string):
        last = m
    if last is None:
        return -2, -1, ""
    return last.start(), last.end(), last.group(0)


def find_unmatched_sgr(string: str) -> str | None:
    reset_pos, _, _ = find_last_reset(string)
    sgr_pos, _, sgr = find_last_sgr(string)
    if sgr_pos > reset_pos:
        return sgr
    else:
        return None


def color_wrap(text: str, width: int = 70) -> list[str]:
    """A variant of wrap that takes SGR codes (somewhat) into account.

    This doesn't actually adjust the length, but makes sure that
    lines that enable some attribues also contain a RESET, and also adds
    that code to the next line
    """
    # TODO we really want to ignore all SGR codes when measuring the width
    lines = wrap(text, width)
    for num, _ in enumerate(lines):
        sgr = find_unmatched_sgr(lines[num])
        if sgr is not None:
            lines[num] += RESET
            if (num + 1) < len(lines):
                lines[num + 1] = sgr + lines[num + 1]
    return lines


def get_weekday_occurrence(day: dt.date) -> tuple[int, int]:
    """Calculate how often this weekday has already occurred in a given month.

    :returns: weekday (0=Monday, ..., 6=Sunday), occurrence
    """
    xthday = 1 + (day.day - 1) // 7
    return day.weekday(), xthday


def get_month_abbr_len() -> int:
    """Calculate the number of characters we need to display the month
    abbreviated name. It depends on the locale.
    """
    return max(len(month_abbr[i]) for i in range(1, 13)) + 1


def localize_strip_tz(dates: list[dt.datetime], timezone: dt.tzinfo) -> Iterator[dt.datetime]:
    """converts a list of dates to timezone, than removes tz info"""
    for one_date in dates:
        if getattr(one_date, "tzinfo", None) is not None:
            one_date = one_date.astimezone(timezone)
            one_date = one_date.replace(tzinfo=None)
        yield one_date


def to_unix_time(dtime: dt.datetime) -> float:
    """convert a datetime object to unix time in UTC (as a float)"""
    if getattr(dtime, "tzinfo", None) is not None:
        dtime = dtime.astimezone(pytz.UTC)
    unix_time = timegm(dtime.timetuple())
    return unix_time


def to_naive_utc(dtime: dt.datetime) -> dt.datetime:
    """convert a datetime object to UTC and than remove the tzinfo, if
    datetime is naive already, return it
    """
    if not hasattr(dtime, "tzinfo") or dtime.tzinfo is None:
        return dtime

    dtime_utc = dtime.astimezone(pytz.UTC)
    dtime_naive = dtime_utc.replace(tzinfo=None)
    return dtime_naive


def is_aware(dtime: dt.datetime) -> bool:
    """test if a datetime instance is timezone aware"""
    return bool(dtime.tzinfo is not None and dtime.tzinfo.utcoffset(dtime) is not None)


def relative_timedelta_str(day: dt.date) -> str:
    """Converts the timespan from `day` to today into a human readable string."""
    days = (day - dt.date.today()).days
    if days < 0:
        direction = "ago"
    else:
        direction = "from now"
    approx = ""
    if abs(days) < 7:
        unit = "day"
        count = abs(days)
    elif abs(days) < 365:
        unit = "week"
        count = int(abs(days) / 7)
        if abs(days) % 7 != 0:
            approx = "~"
    else:
        unit = "year"
        count = int(abs(days) / 365)
        if abs(days) % 365 != 0:
            approx = "~"
    if count > 1:
        unit += "s"

    return f"{approx}{count} {unit} {direction}"


def get_wrapped_text(widget: urwid.AttrMap) -> str:
    return widget.original_widget.get_edit_text()


def human_formatter(format_string, width=None, colors=True):
    """Create a formatter that formats events to be human readable."""

    def fmt(rows):
        single = isinstance(rows, dict)
        if single:
            rows = [rows]
        results = []
        for row in rows:
            if "calendar-color" in row:
                row["calendar-color"] = get_color(row["calendar-color"])

            if "alarms-list" in row and isinstance(row["alarms-list"], list):
                row["alarms-list"] = ", ".join(
                    alarm["description"] + "@" + alarm["delta-formatted"]
                    for alarm in row["alarms-list"]
                )

            s = format_string.format(**row)

            if colors:
                s += style("", reset=True)

            if width:
                results += color_wrap(s, width)
            else:
                results.append(s)
        if single:
            return results[0]
        else:
            return results

    return fmt


CONTENT_ATTRIBUTES = [
    "start",
    "start-long",
    "start-date",
    "start-date-long",
    "start-time",
    "end",
    "end-long",
    "end-date",
    "end-date-long",
    "end-time",
    "duration",
    "start-full",
    "start-long-full",
    "start-date-full",
    "start-date-long-full",
    "start-time-full",
    "end-full",
    "end-long-full",
    "end-date-full",
    "end-date-long-full",
    "end-time-full",
    "duration-full",
    "start-style",
    "end-style",
    "to-style",
    "start-end-time-style",
    "end-necessary",
    "end-necessary-long",
    "repeat-symbol",
    "repeat-pattern",
    "alarms-list",
    "title",
    "organizer",
    "description",
    "location",
    "all-day",
    "categories",
    "uid",
    "url",
    "calendar",
    "calendar-color",
    "status",
    "cancelled",
]


def json_formatter(fields):
    """Create a formatter that formats events in JSON."""

    if len(fields) == 1 and fields[0] == "all":
        fields = CONTENT_ATTRIBUTES

    def fmt(rows):
        single = isinstance(rows, dict)
        if single:
            rows = [rows]

        filtered = []
        for row in rows:
            f = dict(filter(lambda e: e[0] in fields and e[0] in CONTENT_ATTRIBUTES, row.items()))

            if f.get("repeat-symbol", "") != "":
                f["repeat-symbol"] = f["repeat-symbol"].strip()
            if f.get("status", "") != "":
                f["status"] = f["status"].strip()
            if f.get("cancelled", "") != "":
                f["cancelled"] = f["cancelled"].strip()

            filtered.append(f)

        results = [json.dumps(filtered, ensure_ascii=False)]

        if single:
            return results[0]
        else:
            return results

    return fmt


def alarmstr2trigger(alarms: str) -> Iterator[dt.timedelta]:
    """convert a comma separated list of alarm strings to dt.timedelta"""
    for alarm in alarms.split(","):
        alarm = alarm.strip()
        alarm_trig = -1 * guesstimedeltafstr(alarm)
        yield alarm_trig


def str2alarm(alarms: str, description: str) -> Iterator[icalendar.Alarm]:
    """convert a comma separated list of alarm strings to icalendar.Alarm"""
    for alarm_trig in alarmstr2trigger(alarms):
        new_alarm = icalendar.Alarm()
        new_alarm.add("ACTION", "DISPLAY")
        new_alarm.add("TRIGGER", alarm_trig)
        new_alarm.add("DESCRIPTION", description)
        yield new_alarm



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
"""all functions related to terminal display are collected here"""

from itertools import zip_longest
from typing import NamedTuple


class NamedColor(NamedTuple):
    color_index: int
    light: bool


RTEXT = "\x1b[7m"  # reverse
NTEXT = "\x1b[0m"  # normal
BTEXT = "\x1b[1m"  # bold
RESET = "\33[0m"
COLORS: dict[str, NamedColor] = {
    "black": NamedColor(color_index=0, light=False),
    "dark red": NamedColor(color_index=1, light=False),
    "dark green": NamedColor(color_index=2, light=False),
    "brown": NamedColor(color_index=3, light=False),
    "dark blue": NamedColor(color_index=4, light=False),
    "dark magenta": NamedColor(color_index=5, light=False),
    "dark cyan": NamedColor(color_index=6, light=False),
    "white": NamedColor(color_index=7, light=False),
    "light gray": NamedColor(color_index=7, light=True),
    "dark gray": NamedColor(color_index=0, light=True),  # actually light black
    "light red": NamedColor(color_index=1, light=True),
    "light green": NamedColor(color_index=2, light=True),
    "yellow": NamedColor(color_index=3, light=True),
    "light blue": NamedColor(color_index=4, light=True),
    "light magenta": NamedColor(color_index=5, light=True),
    "light cyan": NamedColor(color_index=6, light=True),
}


def get_color(
    fg: str | None = None,
    bg: str | None = None,
    bold_for_light_color: bool = False,
) -> str:
    """convert foreground and/or background color in ANSI color codes

    colors can be a color name from the ANSI color palette (e.g. 'dark green'),
    a number between 0 and 255 (still pass them as a string) or an HTML color in
    the style `#00FF00` or `#ABC`

    :param fg: foreground color
    :param bg: background color
    :returns: ANSI color code
    """

    result = ""
    for colorstring, is_bg in ((fg, False), (bg, True)):
        if colorstring:
            color = "\33["
            if colorstring in COLORS:
                # 16 color palette
                if not is_bg:
                    # foreground color
                    c = 30 + COLORS[colorstring].color_index
                    if COLORS[colorstring].light:
                        if bold_for_light_color:
                            color += "1;"
                        else:
                            c += 60
                else:
                    # background color
                    c = 40 + COLORS[colorstring].color_index
                    if COLORS[colorstring].light:
                        if not bold_for_light_color:
                            c += 60
                color += str(c)
            elif colorstring.isdigit():
                # 256 color palette
                if not is_bg:
                    color += "38;5;" + colorstring
                else:
                    color += "48;5;" + colorstring
            else:
                # HTML-style 24-bit color
                if len(colorstring) == 4:
                    # e.g. #ABC, equivalent to #AABBCC
                    r = int(colorstring[1] * 2, 16)
                    g = int(colorstring[2] * 2, 16)
                    b = int(colorstring[3] * 2, 16)
                else:
                    # e.g. #AABBCC
                    r = int(colorstring[1:3], 16)
                    g = int(colorstring[3:5], 16)
                    b = int(colorstring[5:7], 16)
                if not is_bg:
                    color += f"38;2;{r!s};{g!s};{b!s}"
                else:
                    color += f"48;2;{r!s};{g!s};{b!s}"
            color += "m"
            result += color
    return result


def colored(
    string: str,
    fg: str | None = None,
    bg: str | None = None,
    bold_for_light_color: bool = True,
) -> str:
    """colorize `string` with ANSI color codes

     see get_color for description of `fg`, `bg` and `bold_for_light_color`
    :param string: string to be colorized
    :returns: colorized string
    """
    result = get_color(fg, bg, bold_for_light_color)
    result += string
    if fg or bg:
        result += RESET
    return result


def merge_columns(lcolumn: list[str], rcolumn: list[str], width: int = 25) -> list[str]:
    """merge two lists elementwise together

    Wrap right columns to terminal width.
    If the right list(column) is longer, first lengthen the left one.
    We assume that the left column has width `width`, we cannot find
    out its (real) width automatically since it might contain ANSI
    escape sequences.
    """
    missing = len(rcolumn) - len(lcolumn)
    if missing > 0:
        lcolumn = lcolumn + missing * [width * " "]

    return ["    ".join(one) for one in zip_longest(lcolumn, rcolumn, fillvalue="")]



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import logging
import os

import xdg.BaseDirectory
from configobj import ConfigObj, ConfigObjError, flatten_errors, get_extra_values

from khal import __productname__

try:
    # Available from configobj 5.1.0
    from configobj.validate import Validator
except ModuleNotFoundError:
    from validate import Validator

from collections.abc import Callable

from .exceptions import CannotParseConfigFileError, InvalidSettingsError, NoConfigFile
from .utils import (
    config_checks,
    expand_db_path,
    expand_path,
    get_color_from_vdir,
    get_vdir_type,
    is_color,
    is_timedelta,
    is_timezone,
    monthdisplay_option,
    weeknumber_option,
)

logger = logging.getLogger("khal")
SPECPATH = os.path.join(os.path.dirname(__file__), "khal.spec")


def find_configuration_file() -> str | None:
    """Return the configuration filename.

    Check all the paths for configuration files defined in the XDG Base Directory
    Standard, and return the first one that exists, if any.

    For the common case, this will return ~/.config/khal/config, assuming that it
    exists.
    """

    for dir in xdg.BaseDirectory.xdg_config_dirs:
        path = os.path.join(dir, __productname__, "config")
        if os.path.exists(path):
            return path

    return None


def get_config(
    config_path: str | None = None,
    _get_color_from_vdir: Callable = get_color_from_vdir,
    _get_vdir_type: Callable = get_vdir_type,
) -> ConfigObj:
    """reads the config file, validates it and return a config dict

    :param config_path: path to a custom config file, if none is given the
                        default locations will be searched
    :param _get_color_from_vdir: override get_color_from_vdir for testing purposes
    :param _get_vdir_type: override get_vdir_type for testing purposes
    :returns: configuration
    """
    if config_path is None:
        config_path = find_configuration_file()
    if config_path is None or not os.path.exists(config_path):
        raise NoConfigFile()

    logger.debug(f"using the config file at {config_path}")

    try:
        user_config = ConfigObj(
            config_path,
            configspec=SPECPATH,
            interpolation=False,
            file_error=True,
        )
    except ConfigObjError as error:
        logger.fatal(f"parsing the config file with the following error: {error}")
        logger.fatal(
            "if you recently updated khal, the config file format "
            "might have changed, in that case please consult the "
            "CHANGELOG or other documentation"
        )
        raise CannotParseConfigFileError()

    fdict = {
        "timezone": is_timezone,
        "timedelta": is_timedelta,
        "expand_path": expand_path,
        "expand_db_path": expand_db_path,
        "weeknumbers": weeknumber_option,
        "monthdisplay": monthdisplay_option,
        "color": is_color,
    }
    validator = Validator(fdict)
    results = user_config.validate(validator, preserve_errors=True)

    abort = False
    for section, subsection, config_error in flatten_errors(user_config, results):
        abort = True
        if isinstance(config_error, Exception):
            logger.fatal(f"config error:\nin [{section[0]}] {subsection}: {config_error}")
        else:
            for key in config_error:
                if isinstance(config_error[key], Exception):
                    logger.fatal(
                        "config error:\n"
                        f"in {sectionize(section + [subsection])} {key}: "
                        f"{config_error[key]!s}"
                    )

    if abort or not results:
        raise InvalidSettingsError()

    config_checks(user_config, _get_color_from_vdir, _get_vdir_type)

    extras = get_extra_values(user_config)
    for section, value in extras:
        if section == ():
            logger.warning(f'unknown section "{value}" in config file')
        elif section == ("palette",):
            # we don't validate the palette section, because there is no way to
            # automatically extract valid attributes from the ui module
            continue
        else:
            section = sectionize(section)
            logger.warning(f'unknown key or subsection "{value}" in section "{section}"')
    return user_config


def sectionize(sections: list[str], depth: int = 1) -> str:
    """converts list of string into [list][[of]][[[strings]]]"""
    this_part = depth * "[" + sections[0] + depth * "]"
    if len(sections) > 1:
        return this_part + sectionize(sections[1:], depth=depth + 1)
    else:
        return this_part



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""This module contains the event model with all relevant subclasses and some
helper functions."""

import datetime as dt
import logging
import os
from collections.abc import Callable
from typing import Any

import icalendar
import icalendar.cal
import icalendar.prop
import pytz
from click import style
from pytz.tzinfo import StaticTzInfo

from khal.custom_types import LocaleConfiguration
from khal.exceptions import FatalError
from khal.icalendar import cal_from_ics, delete_instance, invalid_timezone
from khal.parse_datetime import timedelta2str
from khal.plugins import FORMATTERS
from khal.utils import generate_random_uid, is_aware, to_naive_utc, to_unix_time

logger = logging.getLogger("khal")


class Event:
    """base Event class for representing a *recurring instance* of an Event

    (in case of non-recurring events this distinction is irrelevant)
    We keep a copy of this instances start and end time around, because for recurring
    events it might be costly to expand the recursion rules

    important distinction for AllDayEvents:
        all end times are as presented to a user, i.e. an event scheduled for
        only one day will have the same start and end date (even though the
        icalendar standard would have the end date be one day later)
    """

    allday: bool = False

    def __init__(
        self,
        vevents: dict[str, icalendar.Event],
        locale: LocaleConfiguration,
        ref: str | None = None,
        readonly: bool = False,
        href: str | None = None,
        etag: str | None = None,
        calendar: str | None = None,
        color: str | None = None,
        start: dt.datetime | None = None,
        end: dt.datetime | None = None,
        addresses: list[str] | None = None,
    ):
        """
        :param start: start datetime of this event instance
        :param end: end datetime of this event instance
        """
        if self.__class__.__name__ == "Event":
            raise ValueError("do not initialize this class directly")
        if ref is None:
            raise ValueError("ref should not be None")
        self._vevents = vevents
        self.ref = ref
        self._locale = locale
        self.readonly = readonly
        self.href = href
        self.etag = etag
        self.calendar = calendar if calendar else ""
        self.color = color
        self._start: dt.datetime
        self._end: dt.datetime
        self.addresses = addresses if addresses else []

        if start is None:
            self._start = self._vevents[self.ref]["DTSTART"].dt
        else:
            self._start = start
        if end is None:
            try:
                self._end = self._vevents[self.ref]["DTEND"].dt
            except KeyError:
                try:
                    self._end = self._start + self._vevents[self.ref]["DURATION"].dt
                except KeyError:
                    self._end = self._start + dt.timedelta(days=1)
        else:
            self._end = end

    @classmethod
    def _get_type_from_vDDD(cls, start: icalendar.prop.vDDDTypes) -> type:
        """infere the type of the class from the START type of the event"""
        if not isinstance(start.dt, dt.datetime):
            return AllDayEvent
        if "TZID" in start.params or start.dt.tzinfo is not None:
            return LocalizedEvent
        return FloatingEvent

    @classmethod
    def _get_type_from_date(cls, start: dt.datetime) -> type["Event"]:
        if hasattr(start, "tzinfo") and start.tzinfo is not None:
            cls = LocalizedEvent
        elif isinstance(start, dt.datetime):
            cls = FloatingEvent
        elif isinstance(start, dt.date):
            cls = AllDayEvent
        return cls

    @classmethod
    def fromVEvents(
        cls,
        events_list: list[icalendar.Event],
        ref: str | None = None,
        start: dt.datetime | None = None,
        **kwargs,
    ) -> "Event":
        assert isinstance(events_list, list)

        vevents = {}
        for event in events_list:
            if "RECURRENCE-ID" in event:
                if invalid_timezone(event["RECURRENCE-ID"]):
                    default_timezone = kwargs["locale"]["default_timezone"]
                    recur_id = default_timezone.localize(event["RECURRENCE-ID"].dt)
                    ident = str(to_unix_time(recur_id))
                else:
                    ident = str(to_unix_time(event["RECURRENCE-ID"].dt))
                vevents[ident] = event
            else:
                vevents["PROTO"] = event

        if ref is None:
            ref = "PROTO" if ref in vevents else list(vevents.keys())[0]
        try:
            if type(vevents[ref]["DTSTART"].dt) is not type(vevents[ref]["DTEND"].dt):
                raise ValueError("DTSTART and DTEND should be of the same type (datetime or date)")
        except KeyError:
            pass

        if start:
            instcls = cls._get_type_from_date(start)
        else:
            instcls = cls._get_type_from_vDDD(vevents[ref]["DTSTART"])
        return instcls(vevents, ref=ref, start=start, **kwargs)

    @classmethod
    def fromString(cls, ics: str, ref=None, **kwargs) -> "Event":
        calendar_collection = cal_from_ics(ics)
        events = [item for item in calendar_collection.walk() if item.name == "VEVENT"]
        return cls.fromVEvents(events, ref, **kwargs)

    def __lt__(self, other: "Event") -> bool:
        start = self.start_local
        other_start = other.start_local
        if isinstance(start, dt.date) and not isinstance(start, dt.datetime):
            start = dt.datetime.combine(start, dt.time.min)

        if isinstance(other_start, dt.date) and not isinstance(other_start, dt.datetime):
            other_start = dt.datetime.combine(other_start, dt.time.min)

        start = start.replace(tzinfo=None)
        other_start = other_start.replace(tzinfo=None)

        if start == other_start:
            end = self.end_local
            other_end = other.end_local
            if isinstance(end, dt.date) and not isinstance(end, dt.datetime):
                end = dt.datetime.combine(end, dt.time.min)

            if isinstance(other_end, dt.date) and not isinstance(other_end, dt.datetime):
                other_end = dt.datetime.combine(other_end, dt.time.min)

            end = end.replace(tzinfo=None)
            other_end = other_end.replace(tzinfo=None)

            if end == other_end:
                return self.summary < other.summary

            try:
                return end < other_end
            except TypeError:
                raise ValueError(f"Cannot compare events {end} and {other_end}")

        try:
            return start < other_start
        except TypeError:
            raise ValueError(f"Cannot compare events {start} and {other_start}")

    def update_start_end(self, start: dt.datetime, end: dt.datetime) -> None:
        """update start and end time of this event

        calling this on a recurring event will lead to the proto instance
        be set to the new start and end times

        beware, this methods performs some open heart surgery
        """
        if type(start) is not type(end):
            raise ValueError("DTSTART and DTEND should be of the same type (datetime or date)")
        self.__class__ = self._get_type_from_date(start)

        self._vevents[self.ref].pop("DTSTART")
        self._vevents[self.ref].add("DTSTART", start)
        self._start = start
        if not isinstance(end, dt.datetime):
            end = end + dt.timedelta(days=1)
        self._end = end
        if "DTEND" in self._vevents[self.ref]:
            self._vevents[self.ref].pop("DTEND")
            self._vevents[self.ref].add("DTEND", end)
        else:
            self._vevents[self.ref].pop("DURATION")
            self._vevents[self.ref].add("DURATION", end - start)

    @property
    def recurring(self) -> bool:
        try:
            rval = (
                "RRULE" in self._vevents[self.ref]
                or "RECURRENCE-ID" in self._vevents[self.ref]
                or "RDATE" in self._vevents[self.ref]
            )
        except KeyError:
            logger.fatal(
                f"The event at {self.href} might be broken. You might want to "
                "file an issue at https://github.com/pimutils/khal/issues"
            )
            raise
        else:
            return rval

    @property
    def recurpattern(self) -> str:
        if "RRULE" in self._vevents[self.ref]:
            return self._vevents[self.ref]["RRULE"].to_ical().decode("utf-8")
        else:
            return ""

    @property
    def recurobject(self) -> icalendar.vRecur:
        if "RRULE" in self._vevents[self.ref]:
            return self._vevents[self.ref]["RRULE"]
        else:
            return icalendar.vRecur()

    def update_rrule(self, rrule: str) -> None:
        self._vevents["PROTO"].pop("RRULE")
        if rrule is not None:
            self._vevents["PROTO"].add("RRULE", rrule)

    @property
    def recurrence_id(self) -> dt.datetime | str:
        """return the "original" start date of this event (i.e. their recurrence-id)"""
        if self.ref == "PROTO":
            return self.start
        else:
            return dt.datetime.fromtimestamp(int(self.ref), pytz.UTC)

    def increment_sequence(self) -> None:
        """update the SEQUENCE number, call before saving this event"""
        # TODO we might want to do this automatically in raw() everytime
        # the event has changed, this will f*ck up the tests though
        try:
            self._vevents[self.ref]["SEQUENCE"] += 1
        except KeyError:
            self._vevents[self.ref]["SEQUENCE"] = 0

    @property
    def symbol_strings(self) -> dict[str, str]:
        if self._locale["unicode_symbols"]:
            return {
                "recurring": "\N{CLOCKWISE GAPPED CIRCLE ARROW}",
                "alarming": "\N{ALARM CLOCK}",
                "range": "\N{LEFT RIGHT ARROW}",
                "range_end": "\N{RIGHTWARDS ARROW TO BAR}",
                "range_start": "\N{RIGHTWARDS ARROW FROM BAR}",
                "right_arrow": "\N{RIGHTWARDS ARROW}",
                "cancelled": "\N{CROSS MARK}",
                "confirmed": "\N{HEAVY CHECK MARK}",
                "tentative": "?",
                "declined": "\N{CROSS MARK}",
                "accepted": "\N{HEAVY CHECK MARK}",
            }
        else:
            return {
                "recurring": "(R)",
                "alarming": "(A)",
                "range": "<->",
                "range_end": "->|",
                "range_start": "|->",
                "right_arrow": "->",
                "cancelled": "X",
                "confirmed": "V",
                "tentative": "?",
                "declined": "X",
                "accepted": "V",
            }

    @property
    def start_local(self) -> dt.datetime:
        """self.start() localized to local timezone"""
        return self.start

    @property
    def end_local(self) -> dt.datetime:
        """self.end() localized to local timezone"""
        return self.end

    @property
    def start(self) -> dt.datetime:
        """this should return the start date(time) as saved in the event"""
        return self._start

    @property
    def end(self) -> dt.datetime:
        """this should return the end date(time) as saved in the event or
        implicitly defined by start and duration"""
        return self._end

    @property
    def duration(self) -> dt.timedelta:
        try:
            return self._vevents[self.ref]["DURATION"].dt
        except KeyError:
            return self.end - self.start

    @property
    def uid(self) -> str:
        return self._vevents[self.ref]["UID"]

    @property
    def organizer(self) -> str:
        if "ORGANIZER" not in self._vevents[self.ref]:
            return ""
        organizer = self._vevents[self.ref]["ORGANIZER"]
        cn = organizer.params.get("CN", "")
        email = organizer.split(":")[-1]
        if cn:
            return f"{cn} ({email})"
        else:
            return email

    @property
    def url(self) -> str:
        if "URL" not in self._vevents[self.ref]:
            return ""
        return self._vevents[self.ref]["URL"]

    def update_url(self, url: str) -> None:
        if url:
            self._vevents[self.ref]["URL"] = url
        else:
            self._vevents[self.ref].pop("URL")

    @staticmethod
    def _create_calendar() -> icalendar.Calendar:
        """create the calendar"""
        calendar = icalendar.Calendar()
        calendar.add("version", "2.0")
        calendar.add("prodid", "-//PIMUTILS.ORG//NONSGML khal / icalendar //EN")
        return calendar

    @property
    def raw(self) -> str:
        """Creates a VCALENDAR containing VTIMEZONEs"""
        calendar = self._create_calendar()
        tzs = []
        for vevent in self._vevents.values():
            if hasattr(vevent["DTSTART"].dt, "tzinfo") and vevent["DTSTART"].dt.tzinfo is not None:
                tzs.append(vevent["DTSTART"].dt.tzinfo)
            if (
                "DTEND" in vevent
                and hasattr(vevent["DTEND"].dt, "tzinfo")
                and vevent["DTEND"].dt.tzinfo is not None
                and vevent["DTEND"].dt.tzinfo not in tzs
            ):
                tzs.append(vevent["DTEND"].dt.tzinfo)

        for tzinfo in tzs:
            if tzinfo == pytz.UTC:
                continue
            timezone = create_timezone(tzinfo, self.start)
            calendar.add_component(timezone)

        for vevent in self._vevents.values():
            calendar.add_component(vevent)
        return calendar.to_ical().decode("utf-8")

    def export_ics(self, path: str) -> None:
        """export event as ICS"""
        export_path = os.path.expanduser(path)
        with open(export_path, "w") as fh:
            fh.write(self.raw)

    @property
    def summary(self) -> str:
        description = None
        date = self._vevents[self.ref].get("x-birthday", None)
        if date:
            description = "birthday"
        else:
            date = self._vevents[self.ref].get("x-anniversary", None)
            if date:
                description = "anniversary"
            else:
                date = self._vevents[self.ref].get("x-abdate", None)
                if date:
                    description = self._vevents[self.ref].get("x-ablabel", "custom event")

        if date:
            number = self.start_local.year - int(date[:4])
            name = self._vevents[self.ref].get("x-fname", None)
            if int(date[4:6]) == 2 and int(date[6:8]) == 29:
                leap = " (29th of Feb.)"
            else:
                leap = ""
            if (number - 1) % 10 == 0 and number != 11:
                suffix = "st"
            elif (number - 2) % 10 == 0 and number != 12:
                suffix = "nd"
            elif (number - 3) % 10 == 0 and number != 13:
                suffix = "rd"
            else:
                suffix = "th"
            return f"{name}'s {number}{suffix} {description}{leap}"
        else:
            return self._vevents[self.ref].get("SUMMARY", "")

    def update_summary(self, summary: str) -> None:
        self._vevents[self.ref]["SUMMARY"] = summary

    @staticmethod
    def _can_handle_alarm(alarm) -> bool:
        """
        Decides whether we can handle a certain alarm.
        """
        return alarm.get("ACTION") == "DISPLAY" and isinstance(
            alarm.get("TRIGGER").dt, dt.timedelta
        )

    @property
    def alarms(self) -> list[tuple[dt.timedelta, str]]:
        """
        Returns a list of all alarms in th original event that we can handle. Unknown types of
        alarms are ignored.
        """
        return [
            (a.get("TRIGGER").dt, a.get("DESCRIPTION"))
            for a in self._vevents[self.ref].subcomponents
            if a.name == "VALARM" and self._can_handle_alarm(a)
        ]

    def update_alarms(self, alarms: list[tuple[dt.timedelta, str]]) -> None:
        """
        Replaces all alarms in the event that can be handled with the ones provided.
        """
        components = self._vevents[self.ref].subcomponents
        # remove all alarms that we can handle from the subcomponents
        components = [
            c for c in components if not (c.name == "VALARM" and self._can_handle_alarm(c))
        ]
        # add all alarms we could handle from the input
        for alarm in alarms:
            new = icalendar.Alarm()
            new.add("ACTION", "DISPLAY")
            new.add("TRIGGER", alarm[0])
            new.add("DESCRIPTION", alarm[1])
            components.append(new)
        self._vevents[self.ref].subcomponents = components

    @property
    def location(self) -> str:
        return self._vevents[self.ref].get("LOCATION", "")

    def update_location(self, location: str) -> None:
        if location:
            self._vevents[self.ref]["LOCATION"] = location
        else:
            self._vevents[self.ref].pop("LOCATION")

    @property
    def attendees(self) -> str:
        addresses = self._vevents[self.ref].get("ATTENDEE", [])
        if not isinstance(addresses, list):
            addresses = [
                addresses,
            ]
        return ", ".join([address.split(":")[-1] for address in addresses])

    def update_attendees(self, attendees: list[str]):
        assert isinstance(attendees, list)
        attendees = [a.strip().lower() for a in attendees if a != ""]
        if len(attendees) > 0:
            # first check for overlaps in existing attendees.
            # Existing vCalAddress objects will be copied, non-existing
            # vCalAddress objects will be created and appended.
            old_attendees = self._vevents[self.ref].get("ATTENDEE", [])
            unchanged_attendees = []
            vCalAddresses = []
            for attendee in attendees:
                for old_attendee in old_attendees:
                    old_email = old_attendee.lstrip("MAILTO:").lower()
                    if attendee == old_email:
                        vCalAddresses.append(old_attendee)
                        unchanged_attendees.append(attendee)
            for attendee in [a for a in attendees if a not in unchanged_attendees]:
                item = icalendar.prop.vCalAddress(f"MAILTO:{attendee}")
                item.params["ROLE"] = icalendar.prop.vText("REQ-PARTICIPANT")
                item.params["PARTSTAT"] = icalendar.prop.vText("NEEDS-ACTION")
                item.params["CUTYPE"] = icalendar.prop.vText("INDIVIDUAL")
                item.params["RSVP"] = icalendar.prop.vText("TRUE")
                # TODO use khard here to receive full information from email address
                vCalAddresses.append(item)

            self._vevents[self.ref]["ATTENDEE"] = vCalAddresses
        else:
            self._vevents[self.ref].pop("ATTENDEE")

    @property
    def categories(self) -> str:
        try:
            return self._vevents[self.ref].get("CATEGORIES", "").to_ical().decode("utf-8")
        except AttributeError:
            return ""

    def update_categories(self, categories: list[str]) -> None:
        assert isinstance(categories, list)
        categories = [c.strip() for c in categories if c != ""]
        self._vevents[self.ref].pop("CATEGORIES", False)
        if categories:
            self._vevents[self.ref].add("CATEGORIES", categories)

    @property
    def description(self) -> str:
        return self._vevents[self.ref].get("DESCRIPTION", "")

    def update_description(self, description: str):
        if description:
            self._vevents[self.ref]["DESCRIPTION"] = description
        else:
            self._vevents[self.ref].pop("DESCRIPTION")

    @property
    def _recur_str(self) -> str:
        if self.recurring:
            recurstr = " " + self.symbol_strings["recurring"]
        else:
            recurstr = ""
        return recurstr

    @property
    def _alarm_str(self) -> str:
        if self.alarms:
            alarmstr = " " + self.symbol_strings["alarming"]
        else:
            alarmstr = ""
        return alarmstr

    @property
    def _status_str(self) -> str:
        if self.status == "CANCELLED":
            statusstr = self.symbol_strings["cancelled"]
        elif self.status == "TENTATIVE":
            statusstr = self.symbol_strings["tentative"]
        elif self.status == "CONFIRMED":
            statusstr = self.symbol_strings["confirmed"]
        else:
            statusstr = ""
        return statusstr

    @property
    def _partstat_str(self) -> str:
        partstat = self.partstat
        if partstat == "ACCEPTED":
            partstatstr = self.symbol_strings["accepted"]
        elif partstat == "TENTATIVE":
            partstatstr = self.symbol_strings["tentative"]
        elif partstat == "DECLINED":
            partstatstr = self.symbol_strings["declined"]
        else:
            partstatstr = ""
        return partstatstr

    def attributes(
        self,
        relative_to: tuple[dt.date, dt.date] | dt.date,
        env=None,
        colors: bool = True,
    ):
        """
        :param colors: determines if colors codes should be printed or not
        """
        env = env or {}

        attributes: dict[str, Any] = {}
        if isinstance(relative_to, tuple):
            relative_to_start, relative_to_end = relative_to
        else:
            relative_to_start = relative_to_end = relative_to

        if isinstance(relative_to_end, dt.datetime):
            relative_to_end = relative_to_end.date()
        if isinstance(relative_to_start, dt.datetime):
            relative_to_start = relative_to_start.date()

        if isinstance(self.start_local, dt.datetime):
            start_local_datetime = self.start_local
            end_local_datetime = self.end_local
        else:
            start_local_datetime = self._locale["local_timezone"].localize(
                dt.datetime.combine(self.start, dt.time.min)
            )
            end_local_datetime = self._locale["local_timezone"].localize(
                dt.datetime.combine(self.end, dt.time.min)
            )

        day_start = self._locale["local_timezone"].localize(
            dt.datetime.combine(relative_to_start, dt.time.min),
        )
        day_end = self._locale["local_timezone"].localize(
            dt.datetime.combine(relative_to_end, dt.time.max),
        )
        next_day_start = day_start + dt.timedelta(days=1)

        allday = isinstance(self, AllDayEvent)

        attributes["start"] = self.start_local.strftime(self._locale["datetimeformat"])
        attributes["start-long"] = self.start_local.strftime(self._locale["longdatetimeformat"])
        attributes["start-date"] = self.start_local.strftime(self._locale["dateformat"])
        attributes["start-date-long"] = self.start_local.strftime(self._locale["longdateformat"])
        attributes["start-time"] = self.start_local.strftime(self._locale["timeformat"])

        attributes["end"] = self.end_local.strftime(self._locale["datetimeformat"])
        attributes["end-long"] = self.end_local.strftime(self._locale["longdatetimeformat"])
        attributes["end-date"] = self.end_local.strftime(self._locale["dateformat"])
        attributes["end-date-long"] = self.end_local.strftime(self._locale["longdateformat"])
        attributes["end-time"] = self.end_local.strftime(self._locale["timeformat"])

        attributes["duration"] = timedelta2str(self.duration)

        # should only have time attributes at this point (start/end)
        full = {}
        for attr in attributes:
            full[attr + "-full"] = attributes[attr]
        attributes.update(full)

        if allday:
            attributes["start"] = attributes["start-date"]
            attributes["start-long"] = attributes["start-date-long"]
            attributes["start-time"] = ""
            attributes["end"] = attributes["end-date"]
            attributes["end-long"] = attributes["end-date-long"]
            attributes["end-time"] = ""

        tostr = ""
        if self.start_local.timetuple() < relative_to_start.timetuple():
            attributes["start-style"] = self.symbol_strings["right_arrow"]
        elif self.start_local.timetuple() == relative_to_start.timetuple():
            attributes["start-style"] = self.symbol_strings["range_start"]
        else:
            attributes["start-style"] = attributes["start-time"]
            tostr = "-"

        if end_local_datetime in [day_end, next_day_start]:
            if self._locale["timeformat"] == "%H:%M":
                attributes["end-style"] = "24:00"
                tostr = "-"
            else:
                attributes["end-style"] = self.symbol_strings["range_end"]
                tostr = ""
        elif end_local_datetime > day_end:
            attributes["end-style"] = self.symbol_strings["right_arrow"]
            tostr = ""
        else:
            attributes["end-style"] = attributes["end-time"]

        if self.start < self.end:
            attributes["to-style"] = "-"
        else:
            attributes["to-style"] = ""

        if start_local_datetime < day_start and end_local_datetime > day_end:
            attributes["start-end-time-style"] = self.symbol_strings["range"]
        else:
            attributes["start-end-time-style"] = (
                attributes["start-style"] + tostr + attributes["end-style"]
            )

        if allday:
            if self.start == self.end:
                attributes["start-end-time-style"] = ""
            elif self.start == relative_to_start and self.end > relative_to_end:
                attributes["start-end-time-style"] = self.symbol_strings["range_start"]
            elif self.start < relative_to_start and self.end > relative_to_end:
                attributes["start-end-time-style"] = self.symbol_strings["range"]
            elif self.start < relative_to_start and self.end == relative_to_end:
                attributes["start-end-time-style"] = self.symbol_strings["range_end"]
            else:
                attributes["start-end-time-style"] = ""

        if allday:
            attributes["end-necessary"] = ""
            attributes["end-necessary-long"] = ""
            if self.start_local != self.end_local:
                attributes["end-necessary"] = attributes["end-date"]
                attributes["end-necessary-long"] = attributes["end-date-long"]
        else:
            attributes["end-necessary"] = attributes["end-time"]
            attributes["end-necessary-long"] = attributes["end-time"]
            if self.start_local.date() != self.end_local.date():
                attributes["end-necessary"] = attributes["end"]
                attributes["end-necessary-long"] = attributes["end-long"]

        attributes["repeat-symbol"] = self._recur_str
        attributes["repeat-pattern"] = self.recurpattern
        attributes["alarm-symbol"] = self._alarm_str
        attributes["alarms-list"] = [
            {
                "delta": alarm[0].total_seconds(),
                "description": str(alarm[1]),
                "delta-formatted": timedelta2str(alarm[0]),
            }
            for alarm in self.alarms
        ]
        attributes["status-symbol"] = self._status_str
        attributes["partstat-symbol"] = self._partstat_str
        attributes["title"] = self.summary
        attributes["organizer"] = self.organizer.strip()

        formatters = FORMATTERS.values()
        if len(formatters) == 1:
            fmt: Callable[[str], str] = list(formatters)[0]
        else:

            def fmt(s: str) -> str:
                return s.strip()

        attributes["description"] = fmt(self.description)
        attributes["description-separator"] = ""
        if attributes["description"]:
            attributes["description-separator"] = " :: "
        attributes["location"] = self.location.strip()
        attributes["attendees"] = self.attendees
        attributes["all-day"] = str(allday)
        attributes["categories"] = self.categories
        attributes["uid"] = self.uid
        attributes["url"] = self.url
        attributes["url-separator"] = ""
        if attributes["url"]:
            attributes["url-separator"] = " :: "

        if "calendars" in env and self.calendar in env["calendars"]:
            cal = env["calendars"][self.calendar]
            attributes["calendar-color"] = cal.get("color", "")
            attributes["calendar"] = cal.get("displayname", self.calendar)
        else:
            attributes["calendar-color"] = attributes["calendar"] = ""
            attributes["calendar"] = self.calendar

        if colors:
            attributes["reset"] = style("", reset=True)
            attributes["bold"] = style("", bold=True, reset=False)
            for c in ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"]:
                attributes[c] = style("", reset=False, fg=c)
                attributes[c + "-bold"] = style("", reset=False, fg=c, bold=True)
        else:
            attributes["reset"] = attributes["bold"] = ""
            for c in ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"]:
                attributes[c] = attributes[c + "-bold"] = ""

        attributes["nl"] = "\n"
        attributes["tab"] = "\t"
        attributes["bell"] = "\a"

        attributes["status"] = self.status + " " if self.status else ""
        attributes["cancelled"] = "CANCELLED " if self.status == "CANCELLED" else ""
        return attributes

    def duplicate(self) -> "Event":
        """duplicate this event's PROTO event"""
        new_uid = generate_random_uid()
        vevent = self._vevents["PROTO"].copy()
        vevent["SEQUENCE"] = 0
        vevent["UID"] = icalendar.vText(new_uid)
        vevent["SUMMARY"] = icalendar.vText(vevent["SUMMARY"] + " Copy")
        event = self.fromVEvents([vevent], locale=self._locale)
        event.calendar = self.calendar
        event._locale = self._locale
        return event

    def delete_instance(self, instance: dt.datetime) -> None:
        """delete an instance from this event

        we don't check, if that instance is an instance of the recurrence rules
        defined in the event
        """
        assert self.recurring
        delete_instance(self._vevents["PROTO"], instance)

        # in case the instance we want to delete is specified as a RECURRENCE-ID
        # event, we should delete that as well
        to_pop = []
        for key in self._vevents:
            if key == "PROTO":
                continue
            try:
                if self._vevents[key].get("RECURRENCE-ID").dt == instance:
                    to_pop.append(key)
            except TypeError:  # localized/floating datetime mismatch
                continue
        for key in to_pop:
            self._vevents.pop(key)

    @property
    def status(self) -> str:
        return self._vevents[self.ref].get("STATUS", "")

    @property
    def partstat(self) -> str | None:
        for attendee in self._vevents[self.ref].get("ATTENDEE", []):
            for address in self.addresses:
                if attendee == "mailto:" + address:
                    return attendee.params.get("PARTSTAT", "")
        return None


class DatetimeEvent(Event):
    pass


class LocalizedEvent(DatetimeEvent):
    """
    see parent
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        try:
            starttz = getattr(self._vevents[self.ref]["DTSTART"].dt, "tzinfo", None)
        except KeyError:
            msg = (
                f"Cannot understand event {kwargs.get('href')} from "
                f"calendar {kwargs.get('calendar')}, you might want to file an issue at "
                "https://github.com/pimutils/khal/issues"
            )
            logger.fatal(msg)
            raise FatalError(  # because in ikhal you won't see the logger's output
                msg
            )

        if starttz is None:
            starttz = self._locale["default_timezone"]
        try:
            endtz = getattr(self._vevents[self.ref]["DTEND"].dt, "tzinfo", None)
        except KeyError:
            endtz = starttz
        if endtz is None:
            endtz = self._locale["default_timezone"]

        if is_aware(self._start):
            self._start = self._start.astimezone(starttz)
        else:
            self._start = starttz.localize(self._start)

        if is_aware(self._end):
            self._end = self._end.astimezone(endtz)
        else:
            self._end = endtz.localize(self._end)

    @property
    def start_local(self) -> dt.datetime:
        """
        see parent
        """
        return self.start.astimezone(self._locale["local_timezone"])

    @property
    def end_local(self) -> dt.datetime:
        """
        see parent
        """
        return self.end.astimezone(self._locale["local_timezone"])


class FloatingEvent(DatetimeEvent):
    """ """

    allday: bool = False

    @property
    def start_local(self) -> dt.datetime:
        return self._locale["local_timezone"].localize(self.start)

    @property
    def end_local(self) -> dt.datetime:
        return self._locale["local_timezone"].localize(self.end)


class AllDayEvent(Event):
    allday: bool = True

    @property
    def end(self) -> dt.datetime:
        end = super().end
        if end == self.start:
            # https://github.com/pimutils/khal/issues/129
            logger.warning(
                f'{self.href} ("{self.summary}"): The event\'s end '
                "date property contains the same value as the start "
                "date, which is invalid as per RFC 5545. Khal will "
                "assume this is meant to be a single-day event on "
                f"{self.start}"
            )
            end += dt.timedelta(days=1)
        return end - dt.timedelta(days=1)

    @property
    def duration(self) -> dt.timedelta:
        try:
            return self._vevents[self.ref]["DURATION"].dt
        except KeyError:
            return self.end - self.start + dt.timedelta(days=1)


def create_timezone(
    tz: pytz.BaseTzInfo, first_date: dt.datetime | None = None, last_date: dt.datetime | None = None
) -> icalendar.Timezone:
    """
    create an icalendar vtimezone from a pytz.tzinfo object

    :param tz: the timezone
    :param first_date: the very first datetime that needs to be included in the
    transition times, typically the DTSTART value of the (first recurring)
    event
    :param last_date: the last datetime that needs to included, typically the
    end of the (very last) event (of a recursion set)
    :returns: timezone information

    we currently have a problem here:

       pytz.timezones only carry the absolute dates of time zone transitions,
       not their RRULEs. This will a) make for rather bloated VTIMEZONE
       components, especially for long recurring events, b) we'll need to
       specify for which time range this VTIMEZONE should be generated and c)
       will not be valid for recurring events that go into eternity.

    Possible Solutions:

    As this information is not provided by pytz at all, there is no
    easy solution, we'd really need to ship another version of the OLSON DB.

    """
    if isinstance(tz, StaticTzInfo):
        return _create_timezone_static(tz)

    # TODO last_date = None, recurring to infinity

    first_date = dt.datetime.today() if not first_date else to_naive_utc(first_date)
    last_date = first_date + dt.timedelta(days=1) if not last_date else to_naive_utc(last_date)
    timezone = icalendar.Timezone()
    timezone.add("TZID", tz)

    dst = {
        one[2]: "DST" in two.__repr__()
        for one, two in iter(tz._tzinfos.items())  # type: ignore
    }
    bst = {
        one[2]: "BST" in two.__repr__()
        for one, two in iter(tz._tzinfos.items())  # type: ignore
    }

    # looking for the first and last transition time we need to include
    first_num, last_num = 0, len(tz._utc_transition_times) - 1  # type: ignore
    first_tt = tz._utc_transition_times[0]  # type: ignore
    last_tt = tz._utc_transition_times[-1]  # type: ignore
    for num, transtime in enumerate(tz._utc_transition_times):  # type: ignore
        if first_date > transtime > first_tt:
            first_num = num
            first_tt = transtime
        if last_tt > transtime > last_date:
            last_num = num
            last_tt = transtime

    timezones: dict[str, icalendar.Component] = {}
    for num in range(first_num, last_num + 1):
        name = tz._transition_info[num][2]  # type: ignore
        if name in timezones:
            ttime = tz.fromutc(tz._utc_transition_times[num]).replace(tzinfo=None)  # type: ignore
            if "RDATE" in timezones[name]:
                timezones[name]["RDATE"].dts.append(icalendar.prop.vDDDTypes(ttime))
            else:
                timezones[name].add("RDATE", ttime)
            continue

        if dst[name] or bst[name]:
            subcomp = icalendar.TimezoneDaylight()
        else:
            subcomp = icalendar.TimezoneStandard()

        subcomp.add("TZNAME", tz._transition_info[num][2])  # type: ignore
        subcomp.add("DTSTART", tz.fromutc(tz._utc_transition_times[num]).replace(tzinfo=None))  # type: ignore
        subcomp.add("TZOFFSETTO", tz._transition_info[num][0])  # type: ignore
        subcomp.add("TZOFFSETFROM", tz._transition_info[num - 1][0])  # type: ignore
        timezones[name] = subcomp

    for subcomp in timezones.values():
        timezone.add_component(subcomp)

    return timezone


def _create_timezone_static(tz: StaticTzInfo) -> icalendar.Timezone:
    """create an icalendar vtimezone from a StaticTzInfo

    :param tz: the timezone
    :returns: timezone information
    """
    timezone = icalendar.Timezone()
    timezone.add("TZID", tz)
    subcomp = icalendar.TimezoneStandard()
    subcomp.add("TZNAME", tz)
    subcomp.add("DTSTART", dt.datetime(1601, 1, 1))
    subcomp.add("RDATE", dt.datetime(1601, 1, 1))
    subcomp.add("TZOFFSETTO", tz._utcoffset)  # type: ignore
    subcomp.add("TZOFFSETFROM", tz._utcoffset)  # type: ignore
    timezone.add_component(subcomp)
    return timezone



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from khal.khalendar.event import Event

# There should be `khal.khalendar.Event` instead of `Any`
# here and in `Postprocess` below but that results in recursive typing
# which mypy doesn't support until
# https://github.com/python/mypy/issues/731 is implemented.
Render = Callable[
    ["Event", Any],
    str,
]

Postprocess = Callable[[str, "Event", Any], str]



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
import datetime as dt
import logging
import os
import stat
import sys
import textwrap
from shutil import get_terminal_size

import click
import click_log

from . import controllers, plugins
from .cli_utils import (
    _select_one_calendar_callback,
    build_collection,
    calendar_option,
    global_options,
    logger,
    mouse_option,
    multi_calendar_option,
    multi_calendar_select,
    prepare_context,
)
from .exceptions import FatalError
from .plugins import COMMANDS
from .terminal import colored
from .utils import human_formatter, json_formatter

try:
    from setproctitle import setproctitle
except ImportError:

    def setproctitle(_):
        pass


click_log.basic_config("khal")

days_option = click.option("--days", default=None, type=int, help="How many days to include.")
week_option = click.option("--week", "-w", help="Include all events in one week.", is_flag=True)
events_option = click.option("--events", default=None, type=int, help="How many events to include.")
dates_arg = click.argument("dates", nargs=-1)


def time_args(f):
    return dates_arg(events_option(week_option(days_option(f))))


def stringify_conf(conf):
    # since we have only two levels of recursion, a recursive function isn't
    # really worth it
    out = []
    for key, value in conf.items():
        out.append(f"[{key}]")
        for subkey, subvalue in value.items():
            if isinstance(subvalue, dict):
                out.append(f"  [[{subkey}]]")
                for subsubkey, subsubvalue in subvalue.items():
                    out.append(f"    {subsubkey}: {subsubvalue}")
            else:
                out.append(f"  {subkey}: {subvalue}")
    return "\n".join(out)


class _KhalGroup(click.Group):
    def list_commands(self, ctx):
        return super().list_commands(ctx) + list(COMMANDS.keys())

    def get_command(self, ctx, name):
        if name in COMMANDS:
            logger.debug(f"found command {name} as a plugin")
            return COMMANDS[name]
        return super().get_command(ctx, name)


@click.group(cls=_KhalGroup)
@click_log.simple_verbosity_option("khal")
@global_options
@click.pass_context
def cli(ctx, config):
    # setting the process title so it looks nicer in ps
    # shows up as 'khal' under linux and as 'python: khal (python2.7)'
    # under FreeBSD, which is still nicer than the default
    setproctitle("khal")
    if ctx.logfilepath:
        logger = logging.getLogger("khal")
        logger.handlers = [logging.FileHandler(ctx.logfilepath)]
    prepare_context(ctx, config)


@cli.command()
@multi_calendar_option
@click.option("--format", "-f", help=("The format of the events."))
@click.option("--day-format", "-df", help=("The format of the day line."))
@click.option(
    "--once",
    "-o",
    help=("Print each event only once (even if it is repeated or spans multiple days)."),
    is_flag=True,
)
@click.option("--notstarted", help=("Print only events that have not started."), is_flag=True)
@click.argument("DATERANGE", nargs=-1, required=False)
@click.pass_context
def calendar(
    ctx, include_calendar, exclude_calendar, daterange, once, notstarted, format, day_format
):
    """Print calendar with agenda."""
    try:
        rows = controllers.calendar(
            build_collection(
                ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
            ),
            agenda_format=format,
            day_format=day_format,
            once=once,
            notstarted=notstarted,
            daterange=daterange,
            conf=ctx.obj["conf"],
            firstweekday=ctx.obj["conf"]["locale"]["firstweekday"],
            locale=ctx.obj["conf"]["locale"],
            weeknumber=ctx.obj["conf"]["locale"]["weeknumbers"],
            monthdisplay=ctx.obj["conf"]["view"]["monthdisplay"],
            hmethod=ctx.obj["conf"]["highlight_days"]["method"],
            default_color=ctx.obj["conf"]["highlight_days"]["default_color"],
            multiple=ctx.obj["conf"]["highlight_days"]["multiple"],
            multiple_on_overflow=ctx.obj["conf"]["highlight_days"]["multiple_on_overflow"],
            color=ctx.obj["conf"]["highlight_days"]["color"],
            highlight_event_days=ctx.obj["conf"]["default"]["highlight_event_days"],
            bold_for_light_color=ctx.obj["conf"]["view"]["bold_for_light_color"],
            env={"calendars": ctx.obj["conf"]["calendars"]},
        )
        click.echo("\n".join(rows))
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command("list")
@multi_calendar_option
@click.option("--format", "-f", help=("The format of the events."))
@click.option("--day-format", "-df", help=("The format of the day line."))
@click.option(
    "--once",
    "-o",
    is_flag=True,
    help=("Print each event only once (even if it is repeated or spans multiple days)."),
)
@click.option("--notstarted", help=("Print only events that have not started."), is_flag=True)
@click.option("--json", help=("Fields to output in json"), multiple=True)
@click.argument("DATERANGE", nargs=-1, required=False, metavar="[DATETIME [DATETIME | RANGE]]")
@click.pass_context
def klist(
    ctx, include_calendar, exclude_calendar, daterange, once, notstarted, json, format, day_format
):
    """List all events between a start (default: today) and (optional)
    end datetime."""
    enabled_eventformatters = plugins.FORMATTERS
    # TODO: register user given format string as a plugin
    logger.debug(f"{enabled_eventformatters}")
    try:
        event_column = controllers.khal_list(
            build_collection(
                ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
            ),
            agenda_format=format,
            day_format=day_format,
            daterange=daterange,
            once=once,
            notstarted=notstarted,
            conf=ctx.obj["conf"],
            env={"calendars": ctx.obj["conf"]["calendars"]},
            json=json,
        )
        if event_column:
            click.echo("\n".join(event_column))
        else:
            logger.debug("No events found")

    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@calendar_option
@click.option("--interactive", "-i", help=("Add event interactively"), is_flag=True)
@click.option("--location", "-l", help=("The location of the new event."))
@click.option("--categories", "-g", help=("The categories of the new event, comma separated."))
@click.option("--repeat", "-r", help=("Repeat event: daily, weekly, monthly or yearly."))
@click.option("--until", "-u", help=("Stop an event repeating on this date."))
@click.option("--format", "-f", help=("The format to print the event."))
@click.option("--json", help=("Fields to output in json"), multiple=True)
@click.option("--alarms", "-m", help=("Alarm times for the new event as DELTAs comma separated"))
@click.option("--url", help=("URI for the event."))
@click.argument(
    "info", metavar="[START [END | DELTA] [TIMEZONE] [SUMMARY] [:: DESCRIPTION]]", nargs=-1
)
@click.pass_context
def new(
    ctx, calendar, info, location, categories, repeat, until, alarms, url, format, json, interactive
):
    """Create a new event from arguments.

    START and END can be either dates, times or datetimes, please have a
    look at the man page for details.
    Everything that cannot be interpreted as a (date)time or a timezone is
    assumed to be the event's summary, if two colons (::) are present,
    everything behind them is taken as the event's description.
    """
    if not info and not interactive:
        raise click.BadParameter("no details provided, did you mean to use --interactive/-i?")

    calendar = calendar or ctx.obj["conf"]["default"]["default_calendar"]
    if calendar is None:
        if interactive:
            while calendar is None:
                calendar = click.prompt("calendar")
                if calendar == "?":
                    for calendar in ctx.obj["conf"]["calendars"]:
                        click.echo(calendar)
                    calendar = None
                elif calendar not in ctx.obj["conf"]["calendars"]:
                    click.echo("unknown calendar enter ? for list")
                    calendar = None
        else:
            raise click.BadParameter(
                "No default calendar is configured, please provide one explicitly."
            )
    try:
        new_func = controllers.new_from_string
        if interactive:
            new_func = controllers.new_interactive
        new_func(
            build_collection(ctx.obj["conf"], ctx.obj.get("calendar_selection", None)),
            calendar,
            ctx.obj["conf"],
            info=" ".join(info),
            location=location,
            categories=categories,
            repeat=repeat,
            env={"calendars": ctx.obj["conf"]["calendars"]},
            until=until,
            alarms=alarms,
            url=url,
            format=format,
            json=json,
        )
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command("import")
@click.option(
    "--include-calendar",
    "-a",
    help=("The calendar to use."),
    callback=_select_one_calendar_callback,
    multiple=True,
)
@click.option("--batch", help=("do not ask for any confirmation."), is_flag=True)
@click.option("--random_uid", "-r", help=("Select a random uid."), is_flag=True)
@click.argument("ics", type=click.File("rb"), nargs=-1)
@click.option("--format", "-f", help=("The format to print the event."))
@click.pass_context
def import_ics(ctx, ics, include_calendar, batch, random_uid, format):
    """Import events from an .ics file (or stdin).

    If an event with the same UID is already present in the (implicitly)
    selected calendar import will ask before updating (i.e. overwriting)
    that old event with the imported one, unless --batch is given, than it
    will always update. If this behaviour is not desired, use the
    `--random-uid` flag to generate a new, random UID.
    If no calendar is specified (and not `--batch`), you will be asked
    to choose a calendar. You can either enter the number printed behind
    each calendar's name or any unique prefix of a calendar's name.

    """
    if include_calendar:
        ctx.obj["calendar_selection"] = {
            include_calendar,
        }
    collection = build_collection(ctx.obj["conf"], ctx.obj.get("calendar_selection", None))
    if (
        batch
        and len(collection.names) > 1
        and ctx.obj["conf"]["default"]["default_calendar"] is None
    ):
        raise click.UsageError(
            "When using batch import, please specify a calendar to import "
            "into or set the `default_calendar` in the config file."
        )
    rvalue = 0
    # Default to stdin:
    if not ics:
        ics_strs = ((sys.stdin.read(), "stdin"),)
        if not batch:

            def isatty(_file):
                try:
                    return _file.isatty()
                except Exception:
                    return False

            if isatty(sys.stdin) and os.stat("/dev/tty").st_mode & stat.S_IFCHR > 0:
                sys.stdin = open("/dev/tty")
            else:
                logger.warning("/dev/tty does not exist, importing might not work")
    else:
        ics_strs = ((ics_file.read(), ics_file.name) for ics_file in ics)

    for ics_str, filename in ics_strs:
        try:
            controllers.import_ics(
                collection,
                ctx.obj["conf"],
                ics=ics_str,
                batch=batch,
                random_uid=random_uid,
                env={"calendars": ctx.obj["conf"]["calendars"]},
            )
        except FatalError as error:
            logger.debug(error, exc_info=True)
            logger.fatal(f"An error occurred when trying to import the file from {filename}")
            logger.fatal("Events from it will not be available in khal")
            if not batch:
                sys.exit(1)
            rvalue = 1
    sys.exit(rvalue)


@cli.command()
@multi_calendar_option
@mouse_option
@click.pass_context
def interactive(ctx, include_calendar, exclude_calendar, mouse):
    """Interactive UI. Also launchable via `ikhal`."""
    if mouse is not None:
        ctx.obj["conf"]["default"]["enable_mouse"] = mouse
    controllers.interactive(
        build_collection(
            ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
        ),
        ctx.obj["conf"],
    )


@click.command()
@global_options
@multi_calendar_option
@mouse_option
@click.pass_context
def interactive_cli(ctx, config, include_calendar, exclude_calendar, mouse):
    """Interactive UI. Also launchable via `khal interactive`."""
    prepare_context(ctx, config)
    if mouse is not None:
        ctx.obj["conf"]["default"]["enable_mouse"] = mouse
    controllers.interactive(
        build_collection(
            ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
        ),
        ctx.obj["conf"],
    )


@cli.command()
@multi_calendar_option
@click.pass_context
def printcalendars(ctx, include_calendar, exclude_calendar):
    """List all calendars."""
    try:
        click.echo(
            "\n".join(
                build_collection(
                    ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
                ).names
            )
        )
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@click.pass_context
def printformats(ctx):
    """Print a date in all formats.

    Print the date 2013-12-21 21:45 in all configured date(time)
    formats to check if these locale settings are configured to ones
    liking."""
    time = dt.datetime(2013, 12, 21, 21, 45)
    try:
        for strftime_format in [
            "longdatetimeformat",
            "datetimeformat",
            "longdateformat",
            "dateformat",
            "timeformat",
        ]:
            dt_str = time.strftime(ctx.obj["conf"]["locale"][strftime_format])
            click.echo(f"{strftime_format}: {dt_str}")
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@click.argument("ics", type=click.File("rb"), required=False)
@click.option("--format", "-f", help=("The format to print the event."))
@click.pass_context
def printics(ctx, ics, format):
    """Print an ics file (or read from stdin) without importing it.

    Just print the ics file, do nothing else."""
    try:
        if ics:
            ics_str = ics.read()
            name = ics.name
        else:
            ics_str = sys.stdin.read()
            name = "stdin input"
        controllers.print_ics(ctx.obj["conf"], name, ics_str, format)
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@multi_calendar_option
@click.option("--format", "-f", help=("The format of the events."))
@click.option("--json", help=("Fields to output in json"), multiple=True)
@click.argument("search_string")
@click.pass_context
def search(ctx, format, json, search_string, include_calendar, exclude_calendar):
    """Search for events matching SEARCH_STRING.

    For recurring events, only the master event and different overwritten
    events are shown.
    """
    # TODO support for time ranges, location, description etc
    if format is None:
        format = ctx.obj["conf"]["view"]["event_format"]
    try:
        collection = build_collection(
            ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
        )
        events = sorted(collection.search(search_string))
        event_column = []
        term_width, _ = get_terminal_size()
        now = dt.datetime.now()
        env = {"calendars": ctx.obj["conf"]["calendars"]}
        if len(json) == 0:
            formatter = human_formatter(format)
        else:
            formatter = json_formatter(json)
        for event in events:
            desc = textwrap.wrap(formatter(event.attributes(relative_to=now, env=env)), term_width)
            event_column.extend(
                [
                    colored(
                        d,
                        event.color,
                        bold_for_light_color=ctx.obj["conf"]["view"]["bold_for_light_color"],
                    )
                    for d in desc
                ]
            )
        if event_column:
            click.echo("\n".join(event_column))
        else:
            logger.debug("No events found")
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@multi_calendar_option
@click.option("--format", "-f", help=("The format of the events."))
@click.option(
    "--show-past", help=("Show events that have already occurred as options"), is_flag=True
)
@click.argument("search_string", nargs=-1)
@click.pass_context
def edit(ctx, format, search_string, show_past, include_calendar, exclude_calendar):
    """Interactively edit (or delete) events matching the search string."""
    try:
        controllers.edit(
            build_collection(
                ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
            ),
            " ".join(search_string),
            format=format,
            allow_past=show_past,
            locale=ctx.obj["conf"]["locale"],
            conf=ctx.obj["conf"],
        )
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@multi_calendar_option
@click.option("--format", "-f", help=("The format of the events."))
@click.option("--day-format", "-df", help=("The format of the day line."))
@click.option("--notstarted", help=("Print only events that have not started"), is_flag=True)
@click.option("--json", help=("Fields to output in json"), multiple=True)
@click.argument("DATETIME", nargs=-1, required=False, metavar="[[START DATE] TIME | now]")
@click.pass_context
def at(ctx, datetime, notstarted, format, day_format, json, include_calendar, exclude_calendar):
    """Print all events at a specific datetime (defaults to now)."""
    if not datetime:
        datetime = ("now",)
    if format is None:
        format = ctx.obj["conf"]["view"]["event_format"]
    try:
        rows = controllers.khal_list(
            build_collection(
                ctx.obj["conf"], multi_calendar_select(ctx, include_calendar, exclude_calendar)
            ),
            agenda_format=format,
            day_format=day_format,
            datepoint=list(datetime),
            once=True,
            notstarted=notstarted,
            conf=ctx.obj["conf"],
            env={"calendars": ctx.obj["conf"]["calendars"]},
            json=json,
        )
        if rows:
            click.echo("\n".join(rows))
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


@cli.command()
@click.pass_context
def configure(ctx):
    """Helper for initial configuration of khal."""
    from . import configwizard

    try:
        configwizard.configwizard()
    except FatalError as error:
        logger.debug(error, exc_info=True)
        logger.fatal(error)
        sys.exit(1)


main_khal, main_ikhal = cli, interactive_cli



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""this module should contain classes that are specific to ikhal, more
general widgets should go in widgets.py"""

from __future__ import annotations

import logging
import threading
import time
from collections.abc import Callable

import urwid

from .widgets import NColumns

logger = logging.getLogger("khal")


class Pane(urwid.WidgetWrap):
    """An abstract Pane to be used in a Window object."""

    def __init__(self, widget, title=None, description=None) -> None:
        self.widget = widget
        urwid.WidgetWrap.__init__(self, widget)
        self._title = title or ""
        self._description = description or ""
        self.window = None

    @property
    def title(self):
        return self._title

    def selectable(self):
        """mark this widget as selectable"""
        return True

    @property
    def description(self):
        return self._description

    def dialog(self, text: str, buttons: list[tuple[str, Callable]]) -> None:
        """Open a dialog box.

        :param text: Text to appear as the body of the Dialog box
        :param buttons: list of tuples button labels and functions to call
            when the button is pressed
        """
        lines = [urwid.Text(line) for line in text.splitlines()]

        buttons = NColumns(
            [urwid.Button(label, on_press=func) for label, func in buttons],
            outermost=True,
        )
        lines.append(buttons)
        content = urwid.LineBox(urwid.Pile(lines))
        overlay = urwid.Overlay(content, self, "center", ("relative", 70), ("relative", 70), None)
        assert self.window is not None
        self.window.open(overlay)

    def scrollable_dialog(
        self,
        text: str | list[urwid.Text],
        buttons: list[tuple[str, Callable]] | None = None,
        title: str = "Press `ESC` to close this window",
    ) -> None:
        """Open a scrollable dialog box.

        :param text: Text to appear as the body of the Dialog box
        :param buttons: button labels and functions to call when the button is pressed
        """
        if isinstance(text, str):
            body = urwid.ListBox([urwid.Text(line) for line in text.splitlines()])
        else:
            body = urwid.ListBox(text)
        if buttons:
            buttons = NColumns(
                [urwid.Button(label, on_press=func) for label, func in buttons],
                outermost=True,
            )
            content = urwid.LineBox(urwid.Pile([body, ("pack", buttons)]))
        else:
            content = urwid.LineBox(urwid.Pile([body]))

        # put the title on the upper line
        over = urwid.Overlay(
            urwid.Text(" " + title + " "),
            content,
            "center",
            len(title) + 2,
            "top",
            None,
        )
        overlay = urwid.Overlay(
            over, self, "center", ("relative", 70), "middle", ("relative", 70), None
        )
        assert self.window is not None
        self.window.open(overlay)

    def keypress(self, size, key):
        """Handle application-wide key strokes."""
        if key in ["f1", "?"]:
            self.show_keybindings()
        elif key in ["L"]:
            self.show_log()
        else:
            return super().keypress(size, key)

    def show_keybindings(self):
        lines = []
        lines.append(urwid.AttrMap(urwid.Text("  Command              Keys"), "alt header"))
        for command, keys in self._conf["keybindings"].items():
            lines.append(urwid.Text(f"  {command:20} {', '.join(keys)}"))
        self.scrollable_dialog(
            lines,
            title="Press `ESC` to close this window, arrows to scroll",
        )

    def show_log(self):
        self.scrollable_dialog(
            "\n".join(self.window._log),
            title="Press `ESC` to close this window, arrows to scroll",
        )


class Window(urwid.Frame):
    """The main user interface frame.

    A window is a frame which displays a header, a footer and a body.
    The header and the footer are handled by this object, and the body
    is the space where Panes can be displayed.

    Each Pane is an interface to interact with the database in one
    way: list the VCards, edit one VCard, and so on. The Window
    provides a mechanism allowing the panes to chain themselves, and
    to carry data between them.
    """

    def __init__(self, footer="", quit_keys=None) -> None:
        quit_keys = quit_keys or ["q"]
        self._track: list[urwid.Overlay] = []

        header = urwid.AttrWrap(urwid.Text(""), "header")
        footer = urwid.AttrWrap(urwid.Text(footer), "footer")
        urwid.Frame.__init__(
            self,
            urwid.Text(""),
            header=header,
            footer=footer,
        )
        self.update_header()
        self._original_w = None
        self.quit_keys = quit_keys

        def alert(message):
            self.update_header(message, warn=True)

        self._alert_daemon = AlertDaemon(alert)
        self._alert_daemon.start()
        self.alert = self._alert_daemon.alert
        self.loop = None
        self._log: list[str] = []
        self._header_is_warning = False

    def open(self, pane, callback=None):
        """Open a new pane.

        The given pane is added to the track and opened. If the given
        callback is not None, it will be called when this new pane
        will be closed.
        """
        pane.window = self
        self._track.append((pane, callback))
        self._update(pane)

    def backtrack(self, data=None):
        """Unstack the displayed pane.

        The current pane is discarded, and the previous one is
        displayed. If the current pane was opened with a callback,
        this callback is called with the given data (if any) before
        the previous pane gets redrawn.
        """
        _old_pane, cb = self._track.pop()
        if cb:
            cb(data)

        if self._track:
            self._update(self._get_current_pane())
        else:
            raise urwid.ExitMainLoop()

    def is_top_level(self):
        """Is the current pane the top-level one?"""
        return len(self._track) == 1

    def on_key_press(self, key):
        """Handle application-wide key strokes."""
        if key in self.quit_keys:
            self.backtrack()
        elif key == "esc" and not self.is_top_level():
            self.backtrack()
        return key

    def _update(self, pane):
        self.set_body(pane)
        self.clear_header()

    def log(self, record: str):
        self._log.append(record)

    def _get_current_pane(self):
        return self._track[-1][0] if self._track else None

    def clear_header(self):
        """clears header if we are not currently showing a warning"""
        if not self._header_is_warning:
            pane_title = getattr(self._get_current_pane(), "title", "")
            self.header.w.set_text(pane_title)

    def update_header(self, alert=None, warn=False):
        """Update the Windows header line.

        :param alert: additional text to show in header, additionally to
            the current title. If `alert` is a tuple, the first entry must
            be a valid palette entry
        :type alert: str or (palette_entry, str)
        """
        self._header_is_warning = warn
        pane_title = getattr(self._get_current_pane(), "title", None)
        text = []

        for part in (pane_title, alert):
            if part:
                text.append(part)
                text.append(("black", " | "))

        self.header.w.set_text(text[:-1] or "")


class AlertDaemon(threading.Thread):
    def __init__(self, set_msg_func) -> None:
        threading.Thread.__init__(self)
        self._set_msg_func = set_msg_func
        self.daemon = True
        self._start_countdown = threading.Event()

    def alert(self, msg):
        self._set_msg_func(msg)
        self._start_countdown.set()

    def run(self):
        # This is a daemon thread. Since the global namespace is going to
        # vanish on interpreter shutdown, redefine everything from the global
        # namespace here.
        _sleep = time.sleep
        _exception = Exception
        _event = self._start_countdown
        _set_msg = self._set_msg_func

        while True:
            _event.wait()
            _sleep(3)
            try:
                _set_msg(None)
            except _exception:
                pass
            _event.clear()



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
The SQLite backend implementation.
"""

import contextlib
import datetime as dt
import logging
import sqlite3
from collections.abc import Iterable, Iterator
from enum import IntEnum
from os import makedirs, path
from typing import Any

import icalendar
import icalendar.cal
import pytz
from dateutil import parser

from khal import utils
from khal.custom_types import EventTuple, LocaleConfiguration
from khal.icalendar import assert_only_one_uid, cal_from_ics
from khal.icalendar import expand as expand_vevent
from khal.icalendar import sanitize as sanitize_vevent
from khal.icalendar import sort_key as sort_vevent_key

from .exceptions import CouldNotCreateDbDir, NonUniqueUID, OutdatedDbVersionError, UpdateFailed

logger = logging.getLogger("khal")

DB_VERSION = 5  # The current db layout version

RECURRENCE_ID = "RECURRENCE-ID"
THISANDFUTURE = "THISANDFUTURE"
THISANDPRIOR = "THISANDPRIOR"

PROTO = "PROTO"


class EventType(IntEnum):
    DATE = 0
    DATETIME = 1


class SQLiteDb:
    """
    This class should provide a caching database for a calendar, keeping raw
    vevents in one table and allowing to retrieve them by dates (via the help
    of some auxiliary tables)

    :param calendar: names of calendars to select from, those are used as
        additional itentifiers together with event's uids. Each (uid, calendar)
        combination should be unique.
    :param db_path: path where this sqlite database will be saved, if this is
        None, a place according to the XDG specifications will be chosen
    """

    def __init__(
        self,
        calendars: Iterable[str],
        db_path: str | None,
        locale: LocaleConfiguration,
    ) -> None:
        assert db_path is not None
        self.calendars: list[str] = list(calendars)
        self.db_path = path.expanduser(db_path)
        self._create_dbdir()
        self.locale = locale
        self._at_once: bool = False
        self.conn = sqlite3.connect(self.db_path)
        self.cursor = self.conn.cursor()
        self._create_default_tables()
        self._check_calendars_exists()
        self._check_table_version()

    @contextlib.contextmanager
    def at_once(self) -> Iterator["SQLiteDb"]:
        assert not self._at_once
        self._at_once = True
        try:
            yield self
        except Exception:
            raise
        else:
            self.conn.commit()
        finally:
            self._at_once = False

    def _create_dbdir(self) -> None:
        """create the dbdir if it doesn't exist"""
        if self.db_path == ":memory:":
            return None
        dbdir = self.db_path.rsplit("/", 1)[0]
        if not path.isdir(dbdir):
            try:
                logger.debug("trying to create the directory for the db")
                makedirs(dbdir, mode=0o770)
                logger.debug("success")
            except OSError as error:
                logger.critical(f"failed to create {dbdir}: {error}")
                raise CouldNotCreateDbDir()

    def _check_table_version(self) -> None:
        """tests for current db Version
        if the table is still empty, insert db_version
        """
        self.cursor.execute("SELECT version FROM version")
        result = self.cursor.fetchone()
        if result is None:
            self.cursor.execute("INSERT INTO version (version) VALUES (?)", (DB_VERSION,))
            self.conn.commit()
        elif not result[0] == DB_VERSION:
            raise OutdatedDbVersionError(
                str(self.db_path) + " is probably an invalid or outdated database.\n"
                "You should consider removing it and running khal again."
            )

    def _create_default_tables(self) -> None:
        """creates version and calendar tables and inserts table version number"""
        self.cursor.execute("CREATE TABLE IF NOT EXISTS version (version INTEGER)")
        logger.debug("created version table")

        self.cursor.execute("""CREATE TABLE IF NOT EXISTS calendars (
            calendar TEXT NOT NULL UNIQUE,
            resource TEXT NOT NULL,
            ctag TEXT
            )""")
        self.cursor.execute("""CREATE TABLE IF NOT EXISTS events (
                href TEXT NOT NULL,
                calendar TEXT NOT NULL,
                sequence INT,
                etag TEXT,
                item TEXT,
                primary key (href, calendar)
                );""")
        self.cursor.execute("""CREATE TABLE IF NOT EXISTS recs_loc (
            dtstart INT NOT NULL,
            dtend INT NOT NULL,
            href TEXT NOT NULL REFERENCES events( href ),
            rec_inst TEXT NOT NULL,
            ref TEXT NOT NULL,
            dtype INT NOT NULL,
            calendar TEXT NOT NULL,
            primary key (href, rec_inst, calendar)
            );""")
        self.cursor.execute("""CREATE TABLE IF NOT EXISTS recs_float (
            dtstart INT NOT NULL,
            dtend INT NOT NULL,
            href TEXT NOT NULL REFERENCES events( href ),
            rec_inst TEXT NOT NULL,
            ref TEXT NOT NULL,
            dtype INT NOT NULL,
            calendar TEXT NOT NULL,
            primary key (href, rec_inst, calendar)
            );""")
        self.conn.commit()

    def _check_calendars_exists(self) -> None:
        """make sure an entry for the current calendar exists in `calendar`
        table
        """
        for cal in self.calendars:
            self.cursor.execute("""SELECT count(*) FROM calendars WHERE calendar = ?;""", (cal,))
            result = self.cursor.fetchone()

            if result[0] != 0:
                logger.debug(f"tables for calendar {cal} exist")
            else:
                sql_s = "INSERT INTO calendars (calendar, resource) VALUES (?, ?);"
                stuple = (cal, "")
                self.sql_ex(sql_s, stuple)

    def sql_ex(self, statement: str, stuple: tuple) -> list:
        """wrapper for sql statements, does a "fetchall" """
        self.cursor.execute(statement, stuple)
        result = self.cursor.fetchall()
        if not self._at_once:
            self.conn.commit()
        return result

    def update(
        self,
        vevent_str: str,
        href: str,
        etag: str = "",
        calendar: str | None = None,
    ) -> None:
        """insert a new or update an existing event into the db

        This is mostly a wrapper around two SQL statements, doing some cleanup
        before.

        :param vevent_str: event to be inserted or updated.
            We assume that even if it contains more than one VEVENT, that they
            are all part of the same event and all have the same UID
        :param href: href of the card on the server, if this href already
            exists in the db the card gets updated. If no href is given, a
            random href is chosen and it is implied that this card does not yet
            exist on the server, but will be uploaded there on next sync.
        :param etag: the etag of the vcard, if this etag does not match the
            remote etag on next sync, this card will be updated from the server.
            For locally created vcards this should not be set
        """
        assert calendar is not None
        assert href is not None
        ical = cal_from_ics(vevent_str)
        check_for_errors(ical, calendar, href)
        if not assert_only_one_uid(ical):
            logger.warning(
                f"The .ics file at {calendar}/{href} contains multiple UIDs.\n"
                "This should not occur in vdir .ics files.\n"
                "If you didn't edit the file by hand, please report a bug "
                "at https://github.com/pimutils/khal/issues .\n"
                "If you want to import it, please use `khal import FILE`."
            )
            raise NonUniqueUID
        vevents = (
            sanitize_vevent(c, self.locale["default_timezone"], href, calendar)
            for c in ical.walk()
            if c.name == "VEVENT"
        )
        # Need to delete the whole event in case we are updating a
        # recurring event with an event which is either not recurring any
        # more or has EXDATEs, as those would be left in the recursion
        # tables. There are obviously better ways to achieve the same
        # result.
        self.delete(href, calendar=calendar)
        for vevent in sorted(vevents, key=sort_vevent_key):
            check_for_errors(vevent, calendar, href)
            check_support(vevent, href, calendar)
            self._update_impl(vevent, href, calendar)

        sql_s = "INSERT INTO events (item, etag, href, calendar) VALUES (?, ?, ?, ?);"
        stuple = (vevent_str, etag, href, calendar)
        self.sql_ex(sql_s, stuple)

    def update_vcf_dates(
        self, vevent_str: str, href: str, etag: str = "", calendar: str | None = None
    ) -> None:
        """insert events from a vcard into the db

        This is will parse BDAY, ANNIVERSARY, X-ANNIVERSARY and X-ABDATE fields.
        It will also look for any X-ABLABEL fields associated with an X-ABDATE
        and use that in the event description.

        :param vevent_str: contact (vcard) to be parsed.
        :param href: href of the card on the server, if this href already
            exists in the db the card gets updated. If no href is given, a
            random href is chosen and it is implied that this card does not yet
            exist on the server, but will be uploaded there on next sync.
        :param etag: the etag of the vcard, if this etag does not match the
            remote etag on next sync, this card will be updated from the server.
            For locally created vcards this should not be set
        """
        assert calendar is not None
        assert href is not None
        # Delete all event entries for this contact
        self.deletelike(href + "%", calendar=calendar)
        ical = cal_from_ics(vevent_str)
        vcard = ical.walk()[0]
        for key in vcard:
            if key in ["BDAY", "X-ANNIVERSARY", "ANNIVERSARY"] or key.endswith("X-ABDATE"):
                date = vcard[key]
                uuid = vcard.get("UID")
                if isinstance(date, list):
                    logger.warning(
                        f"Vcard {href} in collection {calendar} has more than one "
                        f"{key}, will be skipped and not be available in khal."
                    )
                    continue
                try:
                    if date[0:2] == "--" and date[3] != "-":
                        date = "1900" + date[2:]
                        orig_date = False
                    else:
                        orig_date = True
                    date = parser.parse(date).date()
                except ValueError:
                    logger.warning(f"cannot parse {key} in {href} in collection {calendar}")
                    continue
                if "FN" in vcard:
                    name = vcard["FN"]
                else:
                    vn = vcard["N"]
                    if isinstance(vn, str):  # icalendar < 7.0.0
                        n = vn.split(";")
                        name = " ".join([n[1], n[2], n[0]])
                    else:
                        name = f"{vn.fields.given} {vn.fields.additional} {vn.fields.family}"
                vevent = icalendar.Event()
                vevent.add("dtstart", date)
                vevent.add("dtend", date + dt.timedelta(days=1))
                if date.month == 2 and date.day == 29:  # leap year
                    vevent.add("rrule", {"freq": "YEARLY", "BYYEARDAY": 60})
                else:
                    vevent.add("rrule", {"freq": "YEARLY"})
                description = get_vcard_event_description(vcard, key)
                if orig_date:
                    if key == "BDAY":
                        xtag = "x-birthday"
                    elif key.endswith("ANNIVERSARY"):
                        xtag = "x-anniversary"
                    else:
                        xtag = "x-abdate"
                        vevent.add("x-ablabel", description)
                    vevent.add(xtag, f"{date.year:04}{date.month:02}{date.day:02}")
                    vevent.add("x-fname", name)
                vevent.add("summary", f"{name}'s {description}")
                vevent.add("uid", href + key)
                vevent_str = vevent.to_ical().decode("utf-8")
                self._update_impl(vevent, href + key, calendar)
                sql_s = "INSERT INTO events (item, etag, href, calendar) VALUES (?, ?, ?, ?);"
                stuple = (vevent_str, etag, href + key, calendar)
                try:
                    self.sql_ex(sql_s, stuple)
                except sqlite3.IntegrityError as error:
                    raise UpdateFailed(
                        "Database integrity error creating birthday event "
                        f"on {date} for contact {name} (UID: {uuid}): "
                        f"{error}"
                    )

    def _update_impl(self, vevent: icalendar.cal.Event, href: str, calendar: str) -> None:
        """insert `vevent` into the database

        expand `vevent`'s recurrence rules (if needed) and insert all instance
        in the respective tables
        than insert non-recurring and original recurring (those with an RRULE
        property) events into table `events`
        """
        # TODO FIXME this function is a steaming pile of shit
        rec_id = vevent.get(RECURRENCE_ID)
        if rec_id is None:
            rrange = None
        else:
            rrange = rec_id.params.get("RANGE")

        # testing on datetime.date won't work as datetime is a child of date
        if not isinstance(vevent["DTSTART"].dt, dt.datetime):
            dtype = EventType.DATE
        else:
            dtype = EventType.DATETIME
        if ("TZID" in vevent["DTSTART"].params and dtype == EventType.DATETIME) or getattr(
            vevent["DTSTART"].dt, "tzinfo", None
        ):
            recs_table = "recs_loc"
        else:
            recs_table = "recs_float"

        thisandfuture = rrange == THISANDFUTURE
        if thisandfuture:
            start_shift, duration = calc_shift_deltas(vevent)
            start_shift_seconds = start_shift.days * 3600 * 24 + start_shift.seconds
            duration_seconds = duration.days * 3600 * 24 + duration.seconds

        dtstartend = expand_vevent(vevent, href)
        if not dtstartend:
            # Does this event even have dates? Technically it is possible for
            # events to be empty/non-existent by deleting all their recurrences
            # through EXDATE.
            return

        for dtstart, dtend in dtstartend:
            if dtype == EventType.DATE:
                dbstart = utils.to_unix_time(dtstart)
                dbend = utils.to_unix_time(dtend)
            else:
                dbstart = utils.to_unix_time(dtstart)
                dbend = utils.to_unix_time(dtend)

            if rec_id is not None:
                ref = rec_inst = str(utils.to_unix_time(rec_id.dt))
            else:
                rec_inst = str(dbstart)
                ref = PROTO

            if thisandfuture:
                recs_sql_s = (
                    f"UPDATE {recs_table} SET dtstart = rec_inst + ?, dtend = rec_inst + ?, "
                    "ref = ? WHERE rec_inst >= ? AND href = ? AND calendar = ?;"
                )
                stuple_f = (
                    start_shift_seconds,
                    start_shift_seconds + duration_seconds,
                    ref,
                    rec_inst,
                    href,
                    calendar,
                )
                self.sql_ex(recs_sql_s, stuple_f)
            else:
                recs_sql_s = (
                    f"INSERT OR REPLACE INTO {recs_table} "
                    "(dtstart, dtend, href, ref, dtype, rec_inst, calendar)"
                    "VALUES (?, ?, ?, ?, ?, ?, ?);"
                )
                stuple_n = (dbstart, dbend, href, ref, dtype, rec_inst, calendar)
                self.sql_ex(recs_sql_s, stuple_n)

    def get_ctag(self, calendar: str) -> str | None:
        stuple = (calendar,)
        sql_s = "SELECT ctag FROM calendars WHERE calendar = ?;"
        try:
            ctag = self.sql_ex(sql_s, stuple)[0][0]
            return ctag
        except IndexError:
            return None

    def set_ctag(self, ctag: str, calendar: str) -> None:
        stuple = (
            ctag,
            calendar,
        )
        sql_s = "UPDATE calendars SET ctag = ? WHERE calendar = ?;"
        self.sql_ex(sql_s, stuple)
        self.conn.commit()

    def get_etag(self, href: str, calendar: str) -> str | None:
        """get etag for href

        return: etag
        """
        sql_s = "SELECT etag FROM events WHERE href = ? AND calendar = ?;"
        try:
            etag = self.sql_ex(sql_s, (href, calendar))[0][0]
            return etag
        except IndexError:
            return None

    def delete(self, href: str, etag: Any = None, calendar: str = "") -> None:
        """
        removes the event from the db,

        :param etag: only there for compatibility with vdirsyncer's Storage,
                     we always delete
        """
        assert calendar != ""
        for table in ["recs_loc", "recs_float"]:
            sql_s = f"DELETE FROM {table} WHERE href = ? AND calendar = ?;"
            self.sql_ex(sql_s, (href, calendar))
        sql_s = "DELETE FROM events WHERE href = ? AND calendar = ?;"
        self.sql_ex(sql_s, (href, calendar))

    def deletelike(self, href: str, etag: Any = None, calendar: str = "") -> None:
        """
        removes events from the db that match an SQL 'like' statement,

        :param href: The pattern of hrefs to delete. May contain SQL wildcards
                     like '%'
        :param etag: only there for compatibility with vdirsyncer's Storage,
                     we always delete
        """
        assert calendar != ""
        for table in ["recs_loc", "recs_float"]:
            sql_s = f"DELETE FROM {table} WHERE href LIKE ? AND calendar = ?;"
            self.sql_ex(sql_s, (href, calendar))
        sql_s = "DELETE FROM events WHERE href LIKE ? AND calendar = ?;"
        self.sql_ex(sql_s, (href, calendar))

    def list(self, calendar: str) -> list[tuple[str, str]]:
        """list all events in `calendar`

        used for testing
        :returns: list of (href, etag)
        """
        sql_s = "SELECT href, etag FROM events WHERE calendar = ?;"
        return list(set(self.sql_ex(sql_s, (calendar,))))

    def get_localized_calendars(self, start: dt.datetime, end: dt.datetime) -> Iterable[str]:
        assert start.tzinfo is not None
        assert end.tzinfo is not None
        start_u = utils.to_unix_time(start)
        end_u = utils.to_unix_time(end)
        sql_s = (
            "SELECT events.calendar FROM "
            "recs_loc JOIN events ON "
            "recs_loc.href = events.href AND "
            "recs_loc.calendar = events.calendar WHERE "
            "(dtstart >= ? AND dtstart <= ? OR "
            "dtend > ? AND dtend <= ? OR "
            "dtstart <= ? AND dtend >= ?) AND events.calendar in ({0}) "
            "ORDER BY dtstart"
        )
        stuple = tuple([start_u, end_u, start_u, end_u, start_u, end_u] + list(self.calendars))  # type: ignore
        result = self.sql_ex(sql_s.format(",".join(["?"] * len(self.calendars))), stuple)
        for calendar in result:
            yield calendar[0]  # result is always an iterable, even if getting only one item

    def get_localized(self, start: dt.datetime, end: dt.datetime) -> Iterable[EventTuple]:
        assert start.tzinfo is not None
        assert end.tzinfo is not None
        start_timestamp = utils.to_unix_time(start)
        end_timestamp = utils.to_unix_time(end)
        sql_s = (
            "SELECT item, recs_loc.href, dtstart, dtend, ref, etag, dtype, events.calendar "
            "FROM recs_loc JOIN events ON "
            "recs_loc.href = events.href AND "
            "recs_loc.calendar = events.calendar WHERE "
            "(dtstart >= ? AND dtstart <= ? OR "
            "dtend > ? AND dtend <= ? OR "
            "dtstart <= ? AND dtend >= ?) AND "
            # insert as many "?" as we have configured calendars
            f"events.calendar in ({','.join('?' * len(self.calendars))}) "
            "ORDER BY dtstart"
        )
        stuple = (
            start_timestamp,
            end_timestamp,
            start_timestamp,
            end_timestamp,
            start_timestamp,
            end_timestamp,
        ) + tuple(self.calendars)
        result = self.sql_ex(sql_s, stuple)
        for item, href, start_timestamp, end_timestamp, ref, etag, _dtype, calendar in result:
            start = dt.datetime.fromtimestamp(start_timestamp, pytz.UTC)
            end = dt.datetime.fromtimestamp(end_timestamp, pytz.UTC)
            yield item, href, start, end, ref, etag, calendar

    def get_floating_calendars(self, start: dt.datetime, end: dt.datetime) -> Iterable[str]:
        assert start.tzinfo is None
        assert end.tzinfo is None
        start_u = utils.to_unix_time(start)
        end_u = utils.to_unix_time(end)
        sql_s = (
            "SELECT events.calendar FROM "
            "recs_float JOIN events ON "
            "recs_float.href = events.href AND "
            "recs_float.calendar = events.calendar WHERE "
            "(dtstart >= ? AND dtstart < ? OR "
            "dtend > ? AND dtend <= ? OR "
            "dtstart <= ? AND dtend > ? ) AND events.calendar in ({0}) "
            "ORDER BY dtstart"
        )
        stuple = tuple([start_u, end_u, start_u, end_u, start_u, end_u] + list(self.calendars))  # type: ignore
        result = self.sql_ex(sql_s.format(",".join(["?"] * len(self.calendars))), stuple)
        for calendar in result:
            yield calendar[0]

    def get_floating(self, start: dt.datetime, end: dt.datetime) -> Iterable[EventTuple]:
        """return floating events between `start` and `end`"""
        assert start.tzinfo is None
        assert end.tzinfo is None
        start_dt: dt.datetime | dt.date
        end_dt: dt.datetime | dt.date

        start_u = utils.to_unix_time(start)
        end_u = utils.to_unix_time(end)
        sql_s = (
            "SELECT item, recs_float.href, dtstart, dtend, ref, etag, dtype, events.calendar "
            "FROM recs_float JOIN events ON "
            "recs_float.href = events.href AND "
            "recs_float.calendar = events.calendar WHERE "
            "(dtstart >= ? AND dtstart < ? OR "
            "dtend > ? AND dtend <= ? OR "
            "dtstart <= ? AND dtend > ? ) AND events.calendar in ({0}) "
            "ORDER BY dtstart"
        )
        stuple = tuple([start_u, end_u, start_u, end_u, start_u, end_u] + list(self.calendars))  # type: ignore
        result = self.sql_ex(sql_s.format(",".join(["?"] * len(self.calendars))), stuple)
        for item, href, start_s, end_s, ref, etag, dtype, calendar in result:
            start_dt = dt.datetime.fromtimestamp(start_s, pytz.UTC).replace(tzinfo=None)
            end_dt = dt.datetime.fromtimestamp(end_s, pytz.UTC).replace(tzinfo=None)
            if dtype == EventType.DATE:
                start_dt = start_dt.date()
                end_dt = end_dt.date()
            yield item, href, start_dt, end_dt, ref, etag, calendar

    def get(self, href: str, calendar: str) -> str:
        """returns the ical string matching href and calendar"""
        assert calendar is not None
        sql_s = "SELECT item, etag FROM events WHERE href = ? AND calendar = ?;"
        item, _etag = self.sql_ex(sql_s, (href, calendar))[0]
        return item

    def get_with_etag(self, href: str, calendar: str) -> tuple[str, str]:
        """returns the ical string and its etag matching href and calendar"""
        assert calendar is not None
        sql_s = "SELECT item, etag FROM events WHERE href = ? AND calendar = ?;"
        item, etag = self.sql_ex(sql_s, (href, calendar))[0]
        return item, etag

    def search(
        self, search_string: str
    ) -> Iterable[tuple[str, str, dt.date, dt.date, str, str, str]]:
        """search for events matching `search_string`"""
        sql_s = (
            "SELECT item, recs_loc.href, dtstart, dtend, ref, etag, dtype, events.calendar "
            "FROM recs_loc JOIN events ON "
            "recs_loc.href = events.href AND "
            "recs_loc.calendar = events.calendar "
            "WHERE item LIKE (?) and events.calendar in ({0});"
        )
        stuple = tuple([f"%{search_string}%"] + list(self.calendars))
        result = self.sql_ex(sql_s.format(",".join(["?"] * len(self.calendars))), stuple)
        for item, href, start, end, ref, etag, dtype, calendar in result:
            start = dt.datetime.fromtimestamp(start, pytz.UTC)
            end = dt.datetime.fromtimestamp(end, pytz.UTC)
            if dtype == EventType.DATE:
                start = start.date()
                end = end.date()
            yield item, href, start, end, ref, etag, calendar

        sql_s = (
            "SELECT item, recs_float.href, dtstart, dtend, ref, etag, dtype, events.calendar "
            "FROM recs_float JOIN events ON "
            "recs_float.href = events.href AND "
            "recs_float.calendar = events.calendar "
            "WHERE item LIKE (?) and events.calendar in ({0});"
        )
        stuple = tuple([f"%{search_string}%"] + list(self.calendars))
        result = self.sql_ex(sql_s.format(",".join(["?"] * len(self.calendars))), stuple)
        for item, href, start, end, ref, etag, dtype, calendar in result:
            start = dt.datetime.fromtimestamp(start, pytz.UTC).replace(tzinfo=None)
            end = dt.datetime.fromtimestamp(end, pytz.UTC).replace(tzinfo=None)
            if dtype == EventType.DATE:
                start = start.date()
                end = end.date()
            yield item, href, start, end, ref, etag, calendar


def check_support(vevent: icalendar.cal.Event, href: str, calendar: str) -> None:
    """test if all icalendar features used in this event are supported,
    raise `UpdateFailed` otherwise.
    :param vevent: event to test
    :param href: href of this event, only used for logging
    """
    rec_id = vevent.get(RECURRENCE_ID)

    if rec_id is not None and rec_id.params.get("RANGE") == THISANDPRIOR:
        raise UpdateFailed(
            "The parameter `THISANDPRIOR` is not (and will not be) "
            "supported by khal (as applications supporting the latest "
            f"standard MUST NOT create those. Therefore event {href} from "
            f"calendar {calendar} will not be shown in khal"
        )
    rdate = vevent.get("RDATE")
    if rdate is not None and hasattr(rdate, "params") and rdate.params.get("VALUE") == "PERIOD":
        raise UpdateFailed(
            "`RDATE;VALUE=PERIOD` is currently not supported by khal. "
            f"Therefore event {href} from calendar {calendar} will not be shown in khal.\n"
            "Please post exemplary events (please remove any private data) "
            "to https://github.com/pimutils/khal/issues/152 ."
        )


def check_for_errors(component: icalendar.cal.Component, calendar: str, href: str) -> None:
    """checking if component.errors exists, is not empty and if so warn the user"""
    if hasattr(component, "errors") and component.errors:
        logger.error(f"Errors occurred when parsing {calendar}/{href} for the following reasons:")
        for error in component.errors:
            logger.error(error)
        logger.error("This might lead to this event being shown wrongly or not at all.")


def calc_shift_deltas(vevent: icalendar.Event) -> tuple[dt.timedelta, dt.timedelta]:
    """calculate an event's duration and by how much its start time has shifted
    versus its recurrence-id time

    :param event: an event with a RECURRENCE-ID property
    """
    assert isinstance(vevent, icalendar.Event)  # REMOVE ME
    start_shift = vevent["DTSTART"].dt - vevent["RECURRENCE-ID"].dt
    try:
        duration = vevent["DTEND"].dt - vevent["DTSTART"].dt
    except KeyError:
        duration = vevent["DURATION"].dt
    return start_shift, duration


def get_vcard_event_description(vcard: icalendar.cal.Component, key: str) -> str:
    if key == "BDAY":
        return "birthday"
    elif key.endswith("ANNIVERSARY"):
        return "anniversary"
    elif key.endswith("X-ABDATE"):
        desc_key = key[:-8] + "X-ABLABEL"
        if desc_key in vcard:
            return vcard[desc_key]
        else:
            desc_key = key[:-8] + "X-ABLabel"
            if desc_key in vcard:
                return vcard[desc_key]
            else:
                return "custom event from vcard"
    else:
        return "unknown event from vcard"



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import datetime as dt
import glob
import logging
import os
import pathlib
from collections.abc import Callable, Iterable
from os.path import expanduser, expandvars, join
from typing import Literal

import pytz
import xdg
from tzlocal import get_localzone

try:
    # Available from configobj 5.1.0
    from configobj.validate import VdtValueError
except ModuleNotFoundError:
    from validate import VdtValueError

from khal.khalendar.vdir import CollectionNotFoundError, Vdir
from khal.parse_datetime import guesstimedeltafstr
from khal.terminal import COLORS

from .exceptions import InvalidSettingsError

logger = logging.getLogger("khal")


def is_timezone(tzstring: str | None) -> dt.tzinfo:
    """tries to convert tzstring into a pytz timezone or return local timezone

    raises a VdtvalueError if tzstring is not valid
    """
    if not tzstring:
        # later version of tzlocal return zoneinfo (not pytz) timezones
        # as a lot of our other code can't deal with this yet, we need to force
        # pytz timezones for the time being
        return pytz.timezone(str(get_localzone()))
    try:
        return pytz.timezone(tzstring)
    except pytz.UnknownTimeZoneError:
        raise VdtValueError(f"Unknown timezone {tzstring}")


def is_timedelta(string: str) -> dt.timedelta:
    try:
        return guesstimedeltafstr(string)
    except ValueError:
        raise VdtValueError(f"Invalid timedelta: {string}")


def weeknumber_option(option: str) -> Literal["left", "right", False]:
    """checks if *option* is a valid value

    :param option: the option the user set in the config file
    :returns: 'off', 'left', 'right' or False
    """
    option = option.lower()
    if option == "left":
        return "left"
    elif option == "right":
        return "right"
    elif option in ["off", "false", "0", "no", "none"]:
        return False
    else:
        raise VdtValueError(
            f"Invalid value '{option}' for option 'weeknumber', must be one of "
            "'off', 'left' or 'right'"
        )


def monthdisplay_option(option: str) -> Literal["firstday", "firstfullweek"]:
    """checks if *option* is a valid value

    :param option: the option the user set in the config file
    """
    option = option.lower()
    if option == "firstday":
        return "firstday"
    elif option == "firstfullweek":
        return "firstfullweek"
    else:
        raise VdtValueError(
            f"Invalid value '{option}' for option 'monthdisplay', must be one "
            "of 'firstday' or 'firstfullweek'"
        )


def expand_path(path: str) -> str:
    """expands `~` as well as variable names"""
    return expanduser(expandvars(path))


def expand_db_path(path: str) -> str:
    """expands `~` as well as variable names, defaults to $XDG_CACHE_HOME"""
    if path is None:
        path = join(xdg.BaseDirectory.xdg_cache_home, "khal", "khal.db")
    return expanduser(expandvars(path))


def is_color(color: str) -> str:
    """checks if color represents a valid color

    raises a VdtValueError if color is not valid
    """
    # check if color is
    # 1) the default empty value
    # 2) auto
    # 3) a color name from the 16 color palette
    # 4) a color index from the 256 color palette
    # 5) an HTML-style color code
    if (
        color in ["", "auto"]
        or color in COLORS
        or (color.isdigit() and int(color) >= 0 and int(color) <= 255)
        or (
            color.startswith("#")
            and (len(color) in [4, 7, 9])
            and all(c in "01234567890abcdefABCDEF" for c in color[1:])
        )
    ):
        return color
    raise VdtValueError(color)


def test_default_calendar(config) -> None:
    """test if config['default']['default_calendar'] is set to a sensible
    value
    """
    if config["default"]["default_calendar"] is None:
        pass
    elif config["default"]["default_calendar"] not in config["calendars"]:
        logger.fatal(
            f"in section [default] {config['default']['default_calendar']} is "
            "not valid for 'default_calendar', must be one of "
            f"{config['calendars'].keys()}"
        )
        raise InvalidSettingsError()
    elif config["calendars"][config["default"]["default_calendar"]]["readonly"]:
        logger.fatal("default_calendar may not be read_only!")
        raise InvalidSettingsError()


def get_color_from_vdir(path: str) -> str | None:
    try:
        color = Vdir(path, ".ics").get_meta("color")
    except CollectionNotFoundError:
        color = None
    if color is None or color == "":
        logger.debug(f"Found no or empty file `color` in {path}")
        return None
    color = color.strip()
    try:
        is_color(color)
    except VdtValueError:
        logger.warning(f"Found invalid color `{color}` in {path}color")
        color = None
    return color


def get_unique_name(path: str, names: Iterable[str]) -> str:
    # TODO take care of edge cases, make unique name finding less brain-dead
    try:
        name = Vdir(path, ".ics").get_meta("displayname")
    except CollectionNotFoundError:
        logger.fatal(f"The calendar at `{path}` is not a directory.")
        raise
    if name is None or name == "":
        logger.debug(f"Found no or empty file `displayname` in {path}")
        name = os.path.split(path)[-1]
    if name in names:
        while name in names:
            name = name + "1"
    return name


def get_all_vdirs(expand_path: str) -> Iterable[str]:
    """returns a list of paths, expanded using glob"""
    # FIXME currently returns a list of all directories in path
    # we add an additional / at the end to make sure we are only getting
    # directories
    items = glob.glob(f"{expand_path}/", recursive=True)
    paths = [pathlib.Path(item) for item in sorted(items, key=len, reverse=True)]
    leaves = set()
    parents = set()
    for path in paths:
        if path in parents:
            # we have already seen the current directory as the parent of
            # another directory, so this directory can't be a vdir
            continue
        parents.add(path.parent)
        leaves.add(path)
    # sort to make sure that auto generated names are always identical
    return sorted(os.fspath(path) for path in leaves)


def get_vdir_type(_: str) -> str:
    # TODO implement
    return "calendar"


def validate_palette_entry(attr, definition: str) -> bool:
    if len(definition) not in (2, 3, 5):
        logging.error(
            "Invalid color definition for %s: %s, must be of length, 2, 3, or 5", attr, definition
        )
        return False
    if (definition[0] not in COLORS and definition[0] != "") or (
        definition[1] not in COLORS and definition[1] != ""
    ):
        logging.error(
            "Invalid color definition for %s: %s, must be one of %s",
            attr,
            definition,
            COLORS.keys(),
        )
        return False
    return True


def config_checks(
    config,
    _get_color_from_vdir: Callable = get_color_from_vdir,
    _get_vdir_type: Callable = get_vdir_type,
) -> None:
    """do some tests on the config we cannot do with configobj's validator"""
    # TODO rename or split up, we are also expanding vdirs of type discover
    if len(config["calendars"].keys()) < 1:
        logger.fatal("Found no calendar section in the config file")
        raise InvalidSettingsError()
    config["sqlite"]["path"] = expand_db_path(config["sqlite"]["path"])
    if not config["locale"]["default_timezone"]:
        config["locale"]["default_timezone"] = is_timezone(config["locale"]["default_timezone"])
    if not config["locale"]["local_timezone"]:
        config["locale"]["local_timezone"] = is_timezone(config["locale"]["local_timezone"])

    # expand calendars with type = discover
    # we need a copy of config['calendars'], because we modify config in the body of the loop
    for cname, cconfig in sorted(config["calendars"].items()):
        if not isinstance(config["calendars"][cname], dict):
            logger.fatal("Invalid config file, probably missing calendar sections")
            raise InvalidSettingsError
        if config["calendars"][cname]["type"] == "discover":
            logger.debug(f"discovering calendars in {cconfig['path']}")
            vdirs_discovered = get_all_vdirs(cconfig["path"])
            logger.debug(f"found the following vdirs: {vdirs_discovered}")
            for vdir in vdirs_discovered:
                vdir_config = {
                    "path": vdir,
                    "color": _get_color_from_vdir(vdir) or cconfig.get("color", None),
                    "type": _get_vdir_type(vdir),
                    "readonly": cconfig.get("readonly", False),
                    "priority": 10,
                }
                unique_vdir_name = get_unique_name(vdir, config["calendars"].keys())
                config["calendars"][unique_vdir_name] = vdir_config
            config["calendars"].pop(cname)

    test_default_calendar(config)
    for calendar in config["calendars"]:
        if config["calendars"][calendar]["type"] == "birthdays":
            config["calendars"][calendar]["readonly"] = True
        if config["calendars"][calendar]["color"] == "auto":
            config["calendars"][calendar]["color"] = _get_color_from_vdir(
                config["calendars"][calendar]["path"]
            )

    # check palette settings
    valid_palette = True
    for attr in config.get("palette", []):
        valid_palette = valid_palette and validate_palette_entry(attr, config["palette"][attr])
    if not valid_palette:
        logger.fatal("Invalid palette entry")
        raise InvalidSettingsError()



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from khal.exceptions import Error, FatalError, UnsupportedFeatureError


class UnsupportedRruleExceptionError(UnsupportedFeatureError):
    """we do not support exceptions that do not delete events yet"""

    def __init__(self, message="") -> None:
        x = "This kind of recurrence exception is currently unsupported"
        if message:
            x += f": {message.strip()}"
        UnsupportedFeatureError.__init__(self, x)


class ReadOnlyCalendarError(Error):
    """this calendar is readonly and should not be modifiable from within
    khal"""


class EtagMissmatch(Error):
    """An event is trying to be modified from khal which has also been externally
    modified"""


class OutdatedDbVersionError(FatalError):
    """the db file has an older version and needs to be deleted"""


class CouldNotCreateDbDir(FatalError):
    """the db directory could not be created. Abort."""


class UpdateFailed(Error):
    """could not update the event in the database"""


class DuplicateUid(Error):
    """an event with this UID already exists"""

    existing_href: str | None = None


class NonUniqueUID(Error):
    """the .ics file contains more than one UID"""



# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import datetime as dt
import logging
import os
import re
import textwrap
from collections import OrderedDict, defaultdict
from collections.abc import Callable
from shutil import get_terminal_size

import pytz
from click import confirm, echo, prompt, style

from khal import __productname__, __version__, calendar_display, parse_datetime
from khal.custom_types import (
    EventCreationTypes,
    LocaleConfiguration,
    MonthDisplayType,
    WeekNumbersType,
)
from khal.exceptions import DateTimeParseError, FatalError
from khal.khalendar import CalendarCollection
from khal.khalendar.event import Event
from khal.khalendar.exceptions import DuplicateUid, ReadOnlyCalendarError

from .exceptions import ConfigurationError
from .icalendar import cal_from_ics, split_ics
from .icalendar import sort_key as sort_vevent_key
from .khalendar.vdir import Item
from .parse_datetime import timedelta2str
from .terminal import merge_columns
from .utils import human_formatter, json_formatter

logger = logging.getLogger("khal")


def format_day(day: dt.date, format_string: str, locale, attributes=None):
    if attributes is None:
        attributes = {}

    attributes["date"] = day.strftime(locale["dateformat"])
    attributes["date-long"] = day.strftime(locale["longdateformat"])

    attributes["name"] = parse_datetime.construct_daynames(day)

    colors = {"reset": style("", reset=True), "bold": style("", bold=True, reset=False)}
    for c in ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"]:
        colors[c] = style("", reset=False, fg=c)
        colors[c + "-bold"] = style("", reset=False, fg=c, bold=True)
    attributes.update(colors)
    try:
        return format_string.format(**attributes) + colors["reset"]
    except (KeyError, IndexError):
        raise KeyError(f"cannot format day with: {format_string}")


def calendar(
    collection: CalendarCollection,
    agenda_format=None,
    notstarted: bool = False,
    once=False,
    daterange=None,
    day_format=None,
    locale=None,
    conf=None,
    firstweekday: int = 0,
    weeknumber: WeekNumbersType = False,
    monthdisplay: MonthDisplayType = "firstday",
    hmethod: str = "fg",
    default_color: str = "",
    multiple="",
    multiple_on_overflow: bool = False,
    color="",
    highlight_event_days=0,
    full=False,
    bold_for_light_color: bool = True,
    env=None,
):
    term_width, _ = get_terminal_size()
    lwidth = 27 if conf["locale"]["weeknumbers"] == "right" else 25
    rwidth = term_width - lwidth - 4

    try:
        start, end = start_end_from_daterange(
            daterange,
            locale,
            default_timedelta_date=conf["default"]["timedelta"],
            default_timedelta_datetime=conf["default"]["timedelta"],
        )
    except ValueError as error:
        raise FatalError(error)

    event_column = khal_list(
        collection,
        daterange,
        conf=conf,
        agenda_format=agenda_format,
        day_format=day_format,
        once=once,
        notstarted=notstarted,
        width=rwidth,
        env=env,
    )
    if not event_column:
        event_column = [style("No events", bold=True)]
    month_count = (end.year - start.year) * 12 + end.month - start.month + 1
    calendar_column = calendar_display.vertical_month(
        month=start.month,
        year=start.year,
        count=max(conf["view"]["min_calendar_display"], month_count),
        firstweekday=firstweekday,
        weeknumber=weeknumber,
        monthdisplay=monthdisplay,
        collection=collection,
        hmethod=hmethod,
        default_color=default_color,
        multiple=multiple,
        multiple_on_overflow=multiple_on_overflow,
        color=color,
        highlight_event_days=highlight_event_days,
        locale=locale,
        bold_for_light_color=bold_for_light_color,
    )
    return merge_columns(calendar_column, event_column, width=lwidth)


def start_end_from_daterange(
    daterange: list[str],
    locale: LocaleConfiguration,
    default_timedelta_date: dt.timedelta = dt.timedelta(days=1),
    default_timedelta_datetime: dt.timedelta = dt.timedelta(hours=1),
):
    """
    convert a string description of a daterange into start and end datetime

    if no description is given, return (today, today + default_timedelta_date)

    :param daterange: an iterable of strings that describes `daterange`
    :param locale: locale settings
    """
    if not daterange:
        start = dt.datetime(*dt.date.today().timetuple()[:3])
        end = start + default_timedelta_date
    else:
        start, end, _allday = parse_datetime.guessrangefstr(
            daterange,
            locale,
            default_timedelta_date=default_timedelta_date,
            default_timedelta_datetime=default_timedelta_datetime,
        )
    return start, end


def get_events_between(
    collection: CalendarCollection,
    locale: dict,
    start: dt.datetime,
    end: dt.datetime,
    formatter: Callable,
    notstarted: bool,
    env: dict,
    original_start: dt.datetime,
    seen=None,
    colors: bool = True,
) -> list[str]:
    """returns a list of events scheduled between start and end. Start and end
    are strings or datetimes (of some kind).

    :param collection:
    :param locale:
    :param start: the start datetime
    :param end: the end datetime
    :param formatter: the formatter (see :class:`.utils.human_formatter`)
    :param nostarted: True if each event should start after start (instead of
      be active between start and end)
    :param env: a collection of "static" values like calendar names and color
    :param original_start: start datetime to compare against of notstarted is set
    :param seen:
    :param colors:
    :returns: a list to be printed as the agenda for the given days
    """
    assert not (notstarted and not original_start)

    event_list = []
    if env is None:
        env = {}
    assert start
    assert end
    start_local = locale["local_timezone"].localize(start)
    end_local = locale["local_timezone"].localize(end)

    start = start_local.replace(tzinfo=None)
    end = end_local.replace(tzinfo=None)

    events = sorted(collection.get_localized(start_local, end_local))
    events_float = sorted(collection.get_floating(start, end))
    events = sorted(events + events_float)
    for event in events:
        # yes the logic could be simplified, but I believe it's easier
        # to understand what's going on here this way
        if notstarted:
            if event.allday and event.start < original_start.date():
                continue
            elif not event.allday and event.start_local < original_start:
                continue
        if seen is not None and event.uid in seen:
            continue

        try:
            event_attributes = event.attributes(relative_to=(start, end), env=env, colors=colors)
        except KeyError as error:
            raise FatalError(error)

        event_list.append(event_attributes)
        if seen is not None:
            seen.add(event.uid)

    return formatter(event_list)


def khal_list(
    collection,
    daterange: list[str] | None = None,
    conf: dict | None = None,
    agenda_format=None,
    day_format: str | None = None,
    once=False,
    notstarted: bool = False,
    width: int | None = None,
    env=None,
    datepoint=None,
    json: list | None = None,
):
    """returns a list of all events in `daterange`"""
    assert daterange is not None or datepoint is not None
    assert conf is not None

    # because empty strings are also Falsish
    if agenda_format is None:
        agenda_format = conf["view"]["agenda_event_format"]

    if json:
        formatter = json_formatter(json)
        colors = False
    else:
        formatter = human_formatter(agenda_format, width)
        colors = True

    if daterange is not None:
        if day_format is None:
            day_format = conf["view"]["agenda_day_format"]
        start, end = start_end_from_daterange(
            daterange,
            conf["locale"],
            default_timedelta_date=conf["default"]["timedelta"],
            default_timedelta_datetime=conf["default"]["timedelta"],
        )
        logger.debug(f"Getting all events between {start} and {end}")

    elif datepoint is not None:
        if not datepoint:
            datepoint = ["now"]
        try:
            # hand over a copy of the `datepoint` so error reporting works
            # (we pop from that list in guessdatetimefstr())
            start, allday = parse_datetime.guessdatetimefstr(
                list(datepoint),
                conf["locale"],
                dt.date.today(),
            )
        except (ValueError, IndexError):
            raise FatalError(f"Invalid value of {' '.join(datepoint)} for a datetime")
        if allday:
            logger.debug(f"Got date {start}")
            raise FatalError("Please supply a datetime, not a date.")
        end = start + dt.timedelta(seconds=1)
        if day_format is None:
            day_format = style(
                start.strftime(conf["locale"]["longdatetimeformat"]),
                bold=True,
            )
        logger.debug(f"Getting all events between {start} and {end}")

    event_column: list[str] = []
    once = set() if once else None
    if env is None:
        env = {}

    original_start = conf["locale"]["local_timezone"].localize(start)
    while start < end:
        if start.date() == end.date():
            day_end = end
        else:
            day_end = dt.datetime.combine(start.date(), dt.time.max)
        current_events = get_events_between(
            collection,
            locale=conf["locale"],
            formatter=formatter,
            start=start,
            end=day_end,
            notstarted=notstarted,
            original_start=original_start,
            env=env,
            seen=once,
            colors=colors,
        )
        if day_format and (conf["default"]["show_all_days"] or current_events) and not json:
            if len(event_column) != 0 and conf["view"]["blank_line_before_day"]:
                event_column.append("")
            event_column.append(format_day(start.date(), day_format, conf["locale"]))
        event_column.extend(current_events)
        start = dt.datetime(*start.date().timetuple()[:3]) + dt.timedelta(days=1)

    return event_column


def new_interactive(
    collection,
    calendar_name,
    conf,
    info,
    location=None,
    categories=None,
    repeat=None,
    until=None,
    alarms=None,
    format=None,
    json=None,
    env=None,
    url=None,
):
    info: EventCreationTypes
    try:
        info = parse_datetime.eventinfofstr(
            info,
            conf["locale"],
            default_event_duration=conf["default"]["default_event_duration"],
            default_dayevent_duration=conf["default"]["default_dayevent_duration"],
            adjust_reasonably=True,
        )
    except DateTimeParseError:
        info = {}

    while True:
        summary = info.get("summary")
        if not summary:
            summary = None
        info["summary"] = prompt("summary", default=summary)
        if info["summary"]:
            break
        echo("a summary is required")

    while True:
        range_string = None
        if info.get("dtstart") and info.get("dtend"):
            start_string = info["dtstart"].strftime(conf["locale"]["datetimeformat"])
            end_string = info["dtend"].strftime(conf["locale"]["datetimeformat"])
            range_string = start_string + " " + end_string
        daterange = prompt("datetime range", default=range_string)
        start, end, allday = parse_datetime.guessrangefstr(
            daterange, conf["locale"], adjust_reasonably=True
        )
        info["dtstart"] = start
        info["dtend"] = end
        info["allday"] = allday
        if info["dtstart"] and info["dtend"]:
            break
        echo("invalid datetime range")

    while True:
        tz = info.get("timezone") or conf["locale"]["default_timezone"]
        timezone = prompt("timezone", default=str(tz))
        try:
            tz = pytz.timezone(timezone)
            info["timezone"] = tz
            break
        except pytz.UnknownTimeZoneError:
            echo("unknown timezone")

    info["description"] = prompt("description (or 'None')", default=info.get("description"))
    if info["description"] == "None":
        info["description"] = ""

    info.update(
        {
            "location": location,
            "categories": categories,
            "repeat": repeat,
            "until": until,
            "alarms": alarms,
            "url": url,
        }
    )

    event = new_from_dict(
        info,
        collection,
        conf,
        format=format,
        env=env,
        calendar_name=calendar_name,
        json=json,
    )
    echo("event saved")

    term_width, _ = get_terminal_size()
    edit_event(event, collection, conf["locale"], width=term_width)


def new_from_string(
    collection,
    calendar_name,
    conf,
    info,
    location=None,
    categories=None,
    repeat=None,
    until=None,
    alarms=None,
    url=None,
    format=None,
    json=None,
    env=None,
):
    """construct a new event from a string and add it"""
    info = parse_datetime.eventinfofstr(
        info,
        conf["locale"],
        conf["default"]["default_event_duration"],
        conf["default"]["default_dayevent_duration"],
        adjust_reasonably=True,
    )
    if alarms is None:
        if info["allday"]:
            alarms = timedelta2str(conf["default"]["default_dayevent_alarm"])
        else:
            alarms = timedelta2str(conf["default"]["default_event_alarm"])
    info.update(
        {
            "location": location,
            "categories": categories,
            "repeat": repeat,
            "until": until,
            "alarms": alarms,
            "url": url,
        }
    )
    new_from_dict(
        info,
        collection,
        conf=conf,
        format=format,
        env=env,
        calendar_name=calendar_name,
        json=json,
    )


def new_from_dict(
    event_args: EventCreationTypes,
    collection: CalendarCollection,
    conf,
    calendar_name: str | None = None,
    format=None,
    env=None,
    json=None,
) -> Event:
    """Create a new event from arguments and save in vdirs

    This is a wrapper around CalendarCollection.create_event_from_dict()
    """
    if isinstance(event_args["categories"], str):
        event_args["categories"] = [
            event_args["categories"].strip() for category in event_args["categories"].split(",")
        ]
    try:
        event = collection.create_event_from_dict(event_args, calendar_name=calendar_name)
    except ValueError as error:
        raise FatalError(error)

    try:
        collection.insert(event)
    except ReadOnlyCalendarError:
        raise FatalError(f"ERROR: Cannot modify calendar `{calendar_name}` as it is read-only")

    if conf["default"]["print_new"] == "event":
        if json is None or len(json) == 0:
            if format is None:
                format = conf["view"]["event_format"]
            formatter = human_formatter(format)
        else:
            formatter = json_formatter(json)
        echo(formatter(event.attributes(dt.datetime.now(), env=env)))
    elif conf["default"]["print_new"] == "path":
        assert event.href
        path = os.path.join(collection._calendars[event.calendar]["path"], event.href)
        echo(path)
    return event


def present_options(options, prefix="", sep="  ", width=70):
    option_list = [prefix] if prefix else []
    chars = {}
    for option in options:
        char = options[option]["short"]
        chars[char] = option
        option_list.append(option.replace(char, "[" + char + "]", 1))
    option_string = sep.join(option_list)
    option_string = textwrap.fill(option_string, width)
    char = prompt(option_string)
    if char in chars:
        return chars[char]
    else:
        return None


def edit_event(event, collection, locale, allow_quit=False, width=80):
    options = OrderedDict()
    if allow_quit:
        options["no"] = {"short": "n"}
        options["quit"] = {"short": "q"}
    else:
        options["done"] = {"short": "n"}
    options["summary"] = {"short": "s", "attr": "summary"}
    options["description"] = {"short": "d", "attr": "description", "none": True}
    options["datetime range"] = {"short": "t"}
    options["repeat"] = {"short": "p"}
    options["location"] = {"short": "l", "attr": "location", "none": True}
    options["categories"] = {"short": "c", "attr": "categories", "none": True}
    options["alarm"] = {"short": "a"}
    options["Delete"] = {"short": "D"}
    options["url"] = {"short": "u", "attr": "url", "none": True}
    # some output contains ansi escape sequences (probably only resets)
    # if hitting enter, the output (including the escape sequence) gets parsed
    # and fails the parsing. Therefore we remove ansi escape sequences before
    # parsing.
    ansi = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

    now = dt.datetime.now()

    while True:
        choice = present_options(options, prefix="Edit?", width=width)
        if choice is None:
            echo("unknown choice")
            continue
        if choice == "no":
            return True
        if choice in ["quit", "done"]:
            return False

        edited = False

        if choice == "Delete":
            if confirm("Delete all occurences of event?"):
                collection.delete(event.href, event.etag, event.calendar)
                return True
        elif choice == "datetime range":
            current = human_formatter("{start} {end}")(event.attributes(relative_to=now))
            value = prompt("datetime range", default=current)
            try:
                start, end, allday = parse_datetime.guessrangefstr(ansi.sub("", value), locale)
                if not allday:
                    start = locale["local_timezone"].localize(start)
                    end = locale["local_timezone"].localize(end)
                event.update_start_end(start, end)
                edited = True
            except Exception:
                echo("error parsing range")
        elif choice == "repeat":
            recur = event.recurobject
            freq = recur["freq"] if "freq" in recur else ""
            until = recur["until"] if "until" in recur else ""
            if not freq:
                freq = "None"
            freq = prompt('frequency (or "None")', freq)
            if freq == "None":
                event.update_rrule(None)
            else:
                until = prompt('until (or "None")', until)
                if until == "None":
                    until = None
                rrule = parse_datetime.rrulefstr(freq, until, locale, event.start.tzinfo)
                event.update_rrule(rrule)
            edited = True
        elif choice == "alarm":
            default_alarms = []
            for a in event.alarms:
                s = parse_datetime.timedelta2str(-1 * a[0])
                default_alarms.append(s)

            default = ", ".join(default_alarms)
            if not default:
                default = "None"
            alarm = prompt('alarm (or "None")', default)
            if alarm == "None":
                alarm = ""
            alarm_list = []
            for a in alarm.split(","):
                alarm_trig = -1 * parse_datetime.guesstimedeltafstr(a.strip())
                new_alarm = (alarm_trig, event.description)
                alarm_list += [new_alarm]
            event.update_alarms(alarm_list)
            edited = True
        else:
            attr = options[choice]["attr"]
            default = getattr(event, attr)
            question = choice

            allow_none = False
            if "none" in options[choice] and options[choice]["none"]:
                question += ' (or "None")'
                allow_none = True
                if not default:
                    default = "None"

            value = prompt(question, default)
            if allow_none and value == "None":
                value = ""
            if attr == "categories":
                getattr(event, "update_" + attr)([cat.strip() for cat in value.split(",")])
            else:
                getattr(event, "update_" + attr)(value)
            edited = True

        if edited:
            event.increment_sequence()
            collection.update(event)


def edit(collection, search_string, locale, format=None, allow_past=False, conf=None):
    if conf is not None:
        if format is None:
            format = conf["view"]["event_format"]

    term_width, _ = get_terminal_size()
    now = conf["locale"]["local_timezone"].localize(dt.datetime.now())

    events = sorted(collection.search(search_string))
    for event in events:
        if not allow_past:
            if event.allday and event.end < now.date():
                continue
            elif not event.allday and event.end_local < now:
                continue
        event_text = textwrap.wrap(
            human_formatter(format)(event.attributes(relative_to=now)), term_width
        )
        echo("".join(event_text))
        if not edit_event(event, collection, locale, allow_quit=True, width=term_width):
            return


def interactive(collection, conf):
    """start the interactive user interface"""
    from . import ui

    pane = ui.ClassicView(collection, conf, title="select an event", description="do something")
    ui.start_pane(
        pane,
        pane.cleanup,
        program_info=f"{__productname__} v{__version__}",
        quit_keys=conf["keybindings"]["quit"],
    )


def import_ics(collection, conf, ics, batch=False, random_uid=False, format=None, env=None):
    """
    :param batch: setting this to True will insert without asking for approval,
                  even when an event with the same uid already exists
    :type batch: bool
    :param random_uid: whether to assign a random UID to imported events or not
    :type random_uid: bool
    :param format: the format string to print events with
    :type format: str
    """
    if format is None:
        format = conf["view"]["event_format"]
    try:
        vevents = split_ics(ics, random_uid, conf["locale"]["default_timezone"])
    except Exception as error:
        raise FatalError(error)
    for vevent in vevents:
        import_event(vevent, collection, conf["locale"], batch, format, env)


def import_event(vevent, collection, locale, batch, format=None, env=None):
    """import one event into collection, let user choose the collection

    :type vevent: list of vevents, which can be more than one VEVENT, i.e., the
        same UID, i.e., one "master" event and (optionally) 1+ RECURRENCE-ID events
    :type vevent: list(str)
    """
    # print all sub-events
    if not batch:
        for item in cal_from_ics(vevent).walk():
            if item.name == "VEVENT":
                event = Event.fromVEvents(
                    [item], calendar=collection.default_calendar_name, locale=locale
                )
                echo(human_formatter(format)(event.attributes(dt.datetime.now(), env=env)))

    # get the calendar to insert into
    if not collection.writable_names:
        raise ConfigurationError("No writable calendars found, aborting import.")
    if len(collection.writable_names) == 1:
        calendar_name = collection.writable_names[0]
    elif batch:
        calendar_name = collection.default_calendar_name
    else:
        calendar_names = sorted(collection.writable_names)
        choices = ", ".join([f"{name}({num})" for num, name in enumerate(calendar_names)])
        while True:
            value = prompt(
                f"Which calendar do you want to import to? (unique prefixes are fine)\n{choices}",
                default=collection.default_calendar_name,
            )
            try:
                calendar_name = calendar_names[int(value)]
                break
            except (ValueError, IndexError):
                matches = [x for x in collection.writable_names if x.startswith(value)]
                if len(matches) == 1:
                    calendar_name = matches[0]
                    break
            echo("invalid choice")
    assert calendar_name in collection.writable_names

    if batch or confirm(f"Do you want to import this event into `{calendar_name}`?"):
        try:
            collection.insert(Item(vevent), collection=calendar_name)
        except DuplicateUid:
            if batch or confirm(
                "An event with the same UID already exists. Do you want to update it?"
            ):
                collection.force_update(Item(vevent), collection=calendar_name)
            else:
                logger.warning(f"Not importing event with UID `{event.uid}`")


def print_ics(conf, name, ics, format):
    if format is None:
        format = conf["view"]["event_format"]
    cal = cal_from_ics(ics)
    events = [item for item in cal.walk() if item.name == "VEVENT"]
    events_grouped = defaultdict(list)
    for event in events:
        events_grouped[event["UID"]].append(event)

    vevents = []
    for uid in events_grouped:
        vevents.append(sorted(events_grouped[uid], key=sort_vevent_key))

    echo(f"{len(vevents)} events found in {name}")
    for sub_event in vevents:
        event = Event.fromVEvents(sub_event, locale=conf["locale"])
        echo(human_formatter(format)(event.attributes(dt.datetime.now())))
